import copy
import logging
import os
import warnings
from itertools import chain
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    overload,
)

import numpy as np
import yaml
from sklearn.base import TransformerMixin
from sklearn.preprocessing import StandardScaler
from typing_extensions import Literal

from ertk.config import get_arg_mapping
from ertk.dataset.annotation import read_annotations
from ertk.dataset.features import find_features_file, read_features
from ertk.dataset.utils import get_audio_paths
from ertk.transform import group_transform
from ertk.utils import (
    PathOrStr,
    clip_arrays,
    frame_arrays,
    ordered_intersect,
    pad_arrays,
    transpose_time,
)


class Dataset:
    """Class representing a generic dataset, consisting of a set of
    features and optional partitions and annotations. Has various
    preprocessing methods.

    An annotation is a scalar value for all instances in the dataset.

    A partition is a partition of instances into disjoint groups (e.g.
    speakers). A partition should be complete in that each instance is
    in exactly one group in the partition. Each partition has a
    corresponding annotation with the same name, using the group names
    as categorical annotations.

    Parameters
    ----------
    corpus_info: Pathlike or str, optional
        Path to corpus info in YAML format.
    features: Pathlike or str, optional
        Path to features file, or unique name of features in corpus
        features directory.
    subset: str, optional
        The subset of instances to use.
    """

    _annotations: Dict[str, Dict[str, Any]]
    _corpus: str = ""
    _feature_names: List[str]
    _names: List[str]
    _partitions: Set[str]
    _subset: str = ""
    _subsets: Dict[str, List[str]]
    _x: np.ndarray

    # "Private" vars
    _default_subset: str = ""
    _features: str = ""
    _features_path: Path
    _features_dir: Path
    _subset_paths: Dict[str, Path]

    def __init__(
        self,
        corpus_info: PathOrStr,
        features: Optional[PathOrStr] = None,
        subset: str = "default",
    ):
        self._init()
        self.init_corpus_info(corpus_info)
        self.use_subset(subset)
        if features is not None:
            self.update_features(features)

    def _init(self) -> None:
        self._annotations = {}
        self._feature_names = []
        self._names = []
        self._partitions = set()
        self._subsets = {}
        self._subset_paths = {}
        self._x = np.empty((0, 0), dtype=np.float32)

    @classmethod
    def _copy(cls, inst: "Dataset"):
        new_dataset = Dataset.__new__(cls)
        new_dataset._copy_from_dataset(inst)
        return new_dataset

    def _copy_from_dataset(self, other: "Dataset") -> None:
        self._annotations = copy.deepcopy(other._annotations)
        self._corpus = other._corpus
        self._default_subset = other._default_subset
        self._feature_names = other._feature_names.copy()
        self._features = other._features
        self._features_path = other._features_path
        self._features_dir = other._features_dir
        self._names = other._names.copy()
        self._partitions = other._partitions.copy()
        self._subset = other._subset
        self._subset_paths = other._subset_paths.copy()
        self._subsets = copy.deepcopy(other._subsets)
        self._x = other._x.copy()
        if len(self._x.shape) == 1:
            # Non-contiguous array, so copy each contiguous sub-array
            for i in range(len(self._x)):
                self._x[i] = self._x[i].copy()

    def init_corpus_info(self, path: PathOrStr) -> None:
        """Initialise corpus metadata from YAML.

        Parameters
        ----------
        path: os.Pathlike or str
            The path to a YAML file containing corpus metadata.
        """
        path = Path(path)
        with open(path) as fid:
            doc = yaml.safe_load(fid)
        if not isinstance(doc, dict):
            raise RuntimeError("Corpus info invalid.")
        self._corpus = next(iter(doc))
        corpus_info = doc[self.corpus]

        self._features_dir = path.parent / corpus_info["features_dir"]
        self._default_subset = corpus_info["subsets"]["default"]
        for subset, subset_info in corpus_info["subsets"].items():
            if subset == "default":
                continue
            clips_file = path.parent / f"{subset_info['clips']}.txt"
            self._subset_paths[subset] = clips_file
            self.subsets[subset] = [x.stem for x in get_audio_paths(clips_file)]

        if corpus_info["partitions"]:
            for partition in corpus_info["partitions"]:
                self.update_annotation(
                    partition, path.parent / f"{partition}.csv", dtype=str
                )
        if corpus_info["annotations"]:
            for annotation in corpus_info["annotations"]:
                self.update_annotation(annotation, path.parent / f"{annotation}.csv")

    def _verify_annotations(self, annot_name: Optional[str] = None):
        """Make sure there is a value for each instance for each annotation."""

        def verify_annotation(annot_name: str):
            missing = set(self.names) - self.annotations[annot_name].keys()
            if len(missing) > 0:
                raise RuntimeError(
                    f"Incomplete annotation {annot_name}. These names are missing:"
                    f"{missing}"
                )

        to_verify = self.annotations.keys() if annot_name is None else {annot_name}
        for annot_name in to_verify:
            verify_annotation(annot_name)

    def update_features(self, features: PathOrStr) -> None:
        """Update the features matrix and feature names for this dataset.

        Parameters
        ----------
        features: os.PathLike or str
            Path to a set of features or unique name of features in
            corpus features dir.
        """
        if features == "raw_audio":
            self._features = "raw_audio"
            features = self._subset_paths[self.subset or "all"]
        else:
            features = Path(features)
            if len(features.suffix) == 0:
                features = find_features_file(self._features_dir.glob(f"{features}.*"))
            self._features = features.stem
        self._features_path = features
        data = read_features(features)
        self._feature_names = data.feature_names
        if len(self.names) > 0:
            names = set(self.names)
            if len(names - set(data.names)) > 0:
                raise ValueError(
                    f"Features at {features} don't contain all instances for subset "
                    f"{self.subset}."
                )
            self._x = data.features[[i for i, n in enumerate(data.names) if n in names]]
            self._names = ordered_intersect(data.names, names)
        else:
            self._x = data.features
            self._names = data.names

    def use_subset(self, subset: str = "default") -> None:
        """Use a different subset of the instances.

        Parameters
        ----------
        subset: str
            Name of subset to use. Default is "default" which uses the
            default subset specified in corpus_info.
        """
        if subset == "default":
            subset = self._default_subset
        self._subset = subset
        self._names = self.subsets[subset]
        self._verify_annotations()

    def get_idx_for_names(self, names: Collection[str]) -> np.ndarray:
        """Gets indices of instances corresponding to `names`.

        Parameters
        ----------
        names: collection of str
            The names to get indices for.

        Returns
        -------
        idx: np.ndarray
            The indices corresponding to `names`, in order.
        """
        names = set(names)
        return np.array([i for i, x in enumerate(self.names) if x in names])

    @overload
    def get_idx_for_split(
        self,
        split: Union[str, Dict[str, Union[str, Collection[str]]]],
        return_complement: Literal[False],
    ) -> np.ndarray:
        ...

    @overload
    def get_idx_for_split(
        self,
        split: Union[str, Dict[str, Union[str, Collection[str]]]],
        return_complement: Literal[True],
    ) -> Tuple[np.ndarray, np.ndarray]:
        ...

    def get_idx_for_split(
        self,
        split: Union[str, Dict[str, Union[str, Collection[str]]]],
        return_complement: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Gets indices of instances corresponding to the selection
        given by `split`.

        Parameters
        ----------
        split: str or dict
            Either a string containing the subset to select, groups to
            select, a path to such a config file, or a mapping object
            containing the groups to select.
        return_complement: bool
            If True, return a tuple of `idx`, `comp_idx` where
            `comp_idx` contains the complement indices not in the split.

        Returns
        -------
        idx: np.ndarray
            The corresponding indices.
        comp_idx: np.ndarray, optional
            If `return_complement` is True, this is also returned and
            contains the complement indices.
        """
        if isinstance(split, str):
            if ":" not in split and not Path(split).exists():  # is a subset
                return self.get_idx_for_names(self.subsets[split])
            mapping = get_arg_mapping(split)
        else:
            mapping = split
        indices: Set[int] = set()
        for part_name in mapping:
            group_names = self.get_group_names(part_name)
            group_indices = self.get_group_indices(part_name)
            sel = mapping[part_name]
            sel = [sel] if isinstance(sel, str) else sel
            sel_idx = [group_names.index(x) for x in sel]
            indices.update(np.flatnonzero(np.isin(group_indices, sel_idx)))
        comp_indices = set(range(len(self))) - indices
        if return_complement:
            return np.array(sorted(indices)), np.array(sorted(comp_indices))
        else:
            return np.array(sorted(indices))

    def update_annotation(
        self,
        annot_name: str,
        annotations: Union[PathOrStr, Mapping[str, Any], Sequence[Any]],
        dtype: Optional[Type] = None,
    ) -> None:
        """Add or update an annotation.

        Parameters
        ----------
        annot_name: str
            The name of the annotation.
        annotations: PathLike, str, mapping or sequence
            Annotations to add, similar to update_partition(). If
            PathLike or str, annotations are read from a CSV. If a dict,
            should be of the form {instance: annotation}. If a list,
            should have an annotation for each instance.
        dtype: type, optional
            The type of annotations for reading from CSV file.
        """
        if isinstance(annotations, (os.PathLike, str)):
            annot_dict = read_annotations(annotations, dtype=dtype)
        elif isinstance(annotations, Mapping):
            annot_dict = dict(annotations)
        else:
            annot_dict = {
                k: v for k, v in zip(self.names, annotations) if v is not None
            }

        self._annotations[annot_name] = annot_dict
        if (
            dtype is not None
            and issubclass(dtype, str)
            or isinstance(next(iter(annot_dict.values())), str)
        ):
            self.partitions.add(annot_name)
        self._verify_annotations(annot_name)

    def map_groups(self, part_name: str, mapping: Mapping[str, str]) -> None:
        """Map group names in a partition.

        Parameters
        ----------
        part_name: str
            Name of partition.
        mapping: dict
            Group name mapping.
        """
        groups = self.get_annotations(part_name)
        self.update_annotation(
            part_name, [mapping.get(x, x) for x in groups], dtype=str
        )

    def remove_groups(
        self,
        part_name: str,
        *,
        drop: Collection[str] = None,
        keep: Collection[str] = None,
    ) -> None:
        """Remove instances corresponding to groups from the given
        partition.

        Parameters
        ----------
        part_name: str
            The partition name.
        groups: collection of str
            The groups to remove in given partition.
        """
        drop = set([] if drop is None else drop)
        keep = set([] if keep is None else keep)

        if (len(drop) == 0) == (len(keep) == 0):
            raise ValueError("Exactly one of drop and keep should be non-empty.")

        if len(keep) == 0:
            keep = set(self.get_group_names(part_name)) - drop

        annotation = self.annotations[part_name]
        self.remove_instances(keep=[x for x in self.names if annotation[x] in keep])

    def rename_annotation(self, old_name: str, new_name: str) -> None:
        """Renames an annotation. This is useful for example if you want
        to select a different labelling to use. If an annotation already
        exists with the given new_name, then it is replaced
        destructively.

        Parameters
        ----------
        old_name: str
            The name of the annotation to rename.
        new_name: str
            The new name of the annotation.
        """
        self.annotations[new_name] = self.annotations[old_name]
        del self.annotations[old_name]
        if old_name in self.partitions:
            self.partitions.add(new_name)
            self.partitions.remove(old_name)

    def remove_annotation(self, annot_name: str) -> None:
        """Removes a set of annotations from the dataset. This does not
        remove any instances.

        Parameters
        ----------
        annot_name: str
            Annotation name.
        """
        del self.annotations[annot_name]
        # Use difference_update in case annot_name is not already a partition
        self.partitions.difference_update({annot_name})

    def get_annotations(self, annot_name: str) -> List[Any]:
        """Get a list of annotations, one for each instance currently in
        the dataset.

        Parameters
        ----------
        annot_name: str
            Annotation name.

        Returns
        -------
        A list of values, one for each instance in the datset, in the
        same order they appear in names and x.
        """
        return [self.annotations[annot_name][x] for x in self.names]

    def annotation_type(self, annot_name: str) -> Type:
        return type(next(iter(self.annotations[annot_name].values())))

    def get_group_indices(self, annot_name: str) -> np.ndarray:
        """Gets the group indices (i.e. indices into the groups array)
        for a given partition.

        Parameters
        ----------
        annot_name: str
            The partition name.

        Returns
        -------
        A NumPy array of group indices for each instance in the dataset.
        """
        _, idx = np.unique(self.get_annotations(annot_name), return_inverse=True)
        return idx

    def get_group_counts(self, annot_name: str) -> np.ndarray:
        """Get group counts for a partition.

        Parameters
        ----------
        annot_name: str
            The partition name.

        Returns
        -------
        A NumPy array of counts for the corresponding group in this
        partition.
        """
        _, counts = np.unique(self.get_annotations(annot_name), return_counts=True)
        return counts

    def get_group_names(self, annot_name: str) -> List[str]:
        """Get the names of groups in a partition.

        Parameters
        ----------
        annot_name: str
            Annotation name.
        """
        return list(np.unique(self.get_annotations(annot_name)))

    def update_speakers(
        self, speakers: Union[PathOrStr, Mapping[str, str], Sequence[str]]
    ):
        """Update the speakers for this dataset.

        Parameters
        ----------
        speakers: Path, str, dict or list
            If Path or str, read speaker info from CSV at given path. If
            dict, use speaker dict directly. If list, each speakers[i]
            is the corresponding speaker for instance i.
        """
        self.update_annotation("speaker", speakers, dtype=str)

    def remove_instances(
        self, *, drop: Collection[str] = None, keep: Collection[str] = None
    ):
        """Remove instances from dataset. Recalculate annotations,
        partitions, etc.

        Parameters
        ----------
        drop: collection of str
            Instances to drop.
        keep: collection of str
            Instances to keep. Exactly one of drop and keep should be
            given.
        """
        drop = set([] if drop is None else drop)
        keep = set([] if keep is None else keep)

        if (len(drop) == 0) == (len(keep) == 0):
            raise ValueError("Exactly one of drop and keep should be given.")

        if len(keep) == 0:
            keep = set(self.names) - drop

        idx = self.get_idx_for_names(keep)
        self._names = [self.names[i] for i in idx]
        if len(self._x) > 0:  # To avoid NumPy DeprecationWarning
            self._x = self.x[idx]
        try:
            self._verify_annotations()
        except RuntimeError as e:
            raise RuntimeError(
                "Incomplete annotations after removing instances."
            ) from e

    def clone(self):
        return self.copy()

    def copy(self):
        return self._copy(self)

    def normalise(
        self, partition: str, normaliser: TransformerMixin = StandardScaler()
    ):
        """Transforms the data matrix of this dataset in-place using
        some (offline) normalisation method.

        Parameters
        ----------
        normaliser:
            The transform to apply. Must implement fit_transform().
        partition: str
            The partition to apply in a per-group fashion. If "all",
            then perform global normalisation on all the data.
        """

        if partition == "all":
            groups = np.zeros(len(self.x), dtype=int)
        elif partition == "instance":
            groups = np.arange(len(self.x))
        else:
            groups = self.get_group_indices(partition)
        group_transform(self.x, groups, normaliser, inplace=True)

    def pad_arrays(self, pad: int = 32):
        """Pads each array to the nearest multiple of `pad` greater than
        the array size. Assumes axis 0 of x is time.
        """
        logging.info(f"Padding array lengths to nearest multiple of {pad}.")
        self._x = pad_arrays(self.x, pad=pad)

    def clip_arrays(self, length: int):
        """Clips each array to the specified maximum length."""
        logging.info(f"Clipping arrays to max length {length}.")
        self._x = clip_arrays(self.x, length=length, copy=False)

    def frame_arrays(
        self,
        frame_size: int,
        frame_shift: int,
        max_frames: Optional[int] = None,
    ):
        """Create a sequence of frames from the raw signal."""
        logging.info(f"Framing arrays with size {frame_size} and shift {frame_shift}.")
        self._x = frame_arrays(
            self._x,
            frame_size=frame_size,
            frame_shift=frame_shift,
            max_frames=max_frames,
        )

    def transpose_time(self):
        """Transpose the time and feature axis of each instance."""
        logging.info("Transposing time and feature axis of data.")
        self._x = transpose_time(self._x)

    @property
    def corpus(self) -> str:
        """The corpus this dataset represents."""
        return self._corpus

    @property
    def n_instances(self) -> int:
        """Number of instances in this dataset."""
        return len(self.names)

    @property
    def feature_names(self) -> List[str]:
        """List of feature names."""
        return self._feature_names

    @property
    def n_features(self) -> int:
        """Number of features."""
        return len(self.feature_names)

    @property
    def speaker_names(self) -> List[str]:
        """List of unique speakers in this dataset."""
        return self.get_group_names("speaker")

    @property
    def speaker_counts(self) -> np.ndarray:
        """Number of instances for each speaker."""
        return self.get_group_counts("speaker")

    @property
    def speaker_indices(self) -> np.ndarray:
        """Indices into speakers array of corresponding speaker for each
        instance.
        """
        return self.get_group_indices("speaker")

    @property
    def speakers(self) -> List[str]:
        return self.get_annotations("speaker")

    @property
    def n_speakers(self) -> int:
        return len(self.speaker_names)

    @property
    def partitions(self) -> Set[str]:
        """Partitions in this dataset."""
        return self._partitions

    @property
    def annotations(self) -> Dict[str, Dict[str, Any]]:
        """Each annotation is a mapping from instance name to value."""
        return self._annotations

    @property
    def names(self) -> List[str]:
        """List of instance names."""
        return self._names

    @property
    def subset(self) -> str:
        """Name of clip subset used."""
        return self._subset

    @property
    def subsets(self) -> Dict[str, List[str]]:
        """Dict from subset name to set of clip names."""
        return self._subsets

    @property
    def x(self) -> np.ndarray:
        """The data matrix."""
        return self._x

    def __len__(self) -> int:
        return self.n_instances

    def __str__(self):
        s = f"Corpus: {self.corpus}\n"
        for part in sorted(self.partitions):
            group_names = self.get_group_names(part)
            s += f"partition '{part}' ({len(group_names)} groups)\n"
            if len(group_names) <= 20:
                s += f"\t{dict(zip(group_names, self.get_group_counts(part)))}\n"
        for annot in sorted(self.annotations):
            if annot in self.partitions:
                continue
            s += f"annotation '{annot}'\n"
            annotations = np.array(self.get_annotations(annot))
            s += f"\tmin: {annotations.min():.3f}, max: {annotations.max():.3f}, "
            s += f"mean: {annotations.mean():.3f}\n"
        s += f"{self.n_instances} instances\n"
        s += f"subset: {self.subset}\n"
        s += f"using {self._features} ({self.n_features} features)\n"
        if self.x.dtype == object or len(self.x.shape) == 3:
            lengths = [len(x) for x in self.x]
            s += "Sequences:\n"
            s += f"min length: {np.min(lengths)}\n"
            s += f"mean length: {np.mean(lengths)}\n"
            s += f"max length: {np.max(lengths)}\n"
        return s


class LabelledDataset(Dataset):
    """Class representing a dataset containing discrete labels for each
    instance.
    """

    _label_annot: str

    def __init__(
        self,
        corpus_info: PathOrStr,
        features: Optional[PathOrStr] = None,
        subset: str = "default",
        label: str = "label",
    ):
        super().__init__(corpus_info, features, subset)
        self._label_annot = label

    def _copy_from_dataset(self, other: "Dataset") -> None:
        super()._copy_from_dataset(other)
        self._label_annot = getattr(other, "_label_annot", "label")

    def update_labels(self, labels: Union[PathOrStr, Mapping[str, str], Sequence[str]]):
        self.update_annotation(self.label_annot, labels, dtype=str)

    def map_classes(self, mapping: Mapping[str, str]):
        """Modifies classses based on the mapping in map. Keys not
        corresponding to classes are ignored. The new classes will be
        sorted lexicographically.
        """
        self.map_groups(self.label_annot, mapping)

    def remove_classes(
        self, *, drop: Collection[str] = None, keep: Collection[str] = None
    ):
        """Remove instances with labels not in `keep`."""
        self.remove_groups(self.label_annot, drop=drop, keep=keep)

    @property
    def classes(self) -> List[str]:
        """List of unique class labels."""
        return self.get_group_names(self.label_annot)

    @property
    def n_classes(self) -> int:
        """Number of unique classes."""
        return len(self.classes)

    @property
    def class_counts(self) -> np.ndarray:
        """Number of instances for each class."""
        return self.get_group_counts(self.label_annot)

    @property
    def labels(self) -> List[str]:
        """List of labels for instances."""
        return self.get_annotations(self.label_annot)

    @property
    def label_annot(self) -> str:
        """The annotation used as target label."""
        return self._label_annot

    @label_annot.setter
    def label_annot(self, val: str) -> None:
        if val not in self.annotations:
            raise ValueError(f"'{val}' not in annotations.")
        self._label_annot = val

    @property
    def y(self) -> np.ndarray:
        """The class label array; one label per instance."""
        if self.label_annot in self.partitions:
            return self.get_group_indices(self.label_annot)
        else:
            return np.ndarray(self.get_annotations(self.label_annot))


class CombinedDataset(LabelledDataset):
    """A dataset that joins one or more individual datasets together and
    merges annotations.
    """

    def __init__(self, *datasets: LabelledDataset):
        if len(datasets) == 0:
            raise ValueError("No datasets provided.")
        if len(datasets) == 1:
            self._copy_from_dataset(datasets[0])
            return

        self._init()

        logging.info("Combining " + ", ".join(d.corpus for d in datasets))
        self._corpus = "combined"
        self._names = [f"{d.corpus}_{n}" for d in datasets for n in d.names]
        self._feature_names = datasets[0].feature_names
        self._features = datasets[0]._features
        self._x = np.concatenate([d.x for d in datasets])

        self.update_annotation(
            "corpus", [d.corpus for d in datasets for _ in d.names], dtype=str
        )

        all_speakers = {}
        for d in datasets:
            if "speaker" not in d.annotations:
                all_speakers.update(
                    {f"{d.corpus}_{k}": f"{d.corpus}_unknown" for k in d.names}
                )
            else:
                speakers = d.annotations["speaker"]
                all_speakers.update(
                    {f"{d.corpus}_{k}": f"{d.corpus}_{v}" for k, v in speakers.items()}
                )
        self.update_speakers(all_speakers)

        # TODO: handle case when only some datasets have a given annotation?
        all_annotations = set(chain(*(d.annotations for d in datasets)))
        all_annotations -= {"speaker"}  # assumed to be per-dataset
        common_annotations = {
            x for x in all_annotations if all(x in d.annotations for d in datasets)
        }
        logging.info(f"All annotations: {all_annotations}")
        logging.info(f"Common annotations: {common_annotations}")
        for annot_name in common_annotations:
            combined_annot = {}
            for dataset in datasets:
                annotations = dataset.annotations[annot_name]
                combined_annot.update(
                    {f"{dataset.corpus}_{k}": v for k, v in annotations.items()}
                )
            self.update_annotation(annot_name, combined_annot)

        # Not common categorical annotations
        for annot_name in all_annotations - common_annotations:
            if not any(annot_name in d.partitions for d in datasets):
                continue
            combined_annot = {}
            for d in datasets:
                if annot_name not in d.annotations:
                    combined_annot.update(
                        {f"{d.corpus}_{k}": "unknown" for k in d.names}
                    )
                else:
                    annotations = d.annotations[annot_name]
                    combined_annot.update(
                        {f"{d.corpus}_{k}": v for k, v in annotations.items()}
                    )
            self.update_annotation(annot_name, combined_annot)

        if any(d.label_annot != datasets[0].label_annot for d in datasets):
            # Try and unify label annotations with different names
            warnings.warn(
                "label annotation differs between some datasets, attempting to unify."
            )
            self._label_annot = "_combined_labels"
            combined_labels = {}
            for d in datasets:
                for name in d.names:
                    label = d.annotations[d.label_annot][name]
                    combined_labels[f"{d.corpus}_{name}"] = label
            self.update_labels(combined_labels)
        else:
            self._label_annot = next(d.label_annot for d in datasets)

    @property
    def corpus_names(self) -> Sequence[str]:
        """List of corpora in this CombinedDataset."""
        return self.get_group_names("corpus")

    @property
    def corpus_indices(self) -> np.ndarray:
        """Indices into corpora list of corresponding corpus for each
        instance.
        """
        return self.get_group_indices("corpus")


def load_multiple(
    corpus_files: Iterable[PathOrStr],
    features: str,
    subsets: Union[str, Mapping[str, str]] = "default",
    label: Union[str, Mapping[str, str]] = "label",
) -> CombinedDataset:
    """Load one or more datasets with the given features.

    Parameters
    ----------
    corpus_files: iterable
        The corpus description YAML files to load.
    features: str
        A common set of features to load. This will be found in the
        features directory corresponding to each corpus.
    subsets: str or dict
        A subset name common to all datasets (e.g. "all", "default") or
        a mapping from dataset name to subset name.
    """
    corpus_files = list(corpus_files)
    if len(corpus_files) == 0:
        raise RuntimeError("No corpus metadata files given.")
    datasets = []
    for file in corpus_files:
        logging.info(f"Loading {file}")
        dataset = LabelledDataset(file, subset="all")
        if isinstance(label, str):
            dataset.label_annot = label
        else:
            dataset.label_annot = label.get(dataset.corpus, "label")
        if isinstance(subsets, str):
            dataset.use_subset(subsets)
        else:
            dataset.use_subset(subsets.get(dataset.corpus, "default"))
        dataset.update_features(features)
        datasets.append(dataset)
    return CombinedDataset(*datasets)



import liblinear.liblinearutil as liblinearutil
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin
from sklearn.linear_model import LogisticRegression

from ertk.sklearn.utils import OneVsRestClassifier

_EPS = 1e-15


class BaseBinaryMTFL(BaseEstimator):
    """Linear regressor based on MTFL. Code adapted from
    https://github.com/argyriou/multi_task_learning

    Parameters
    ----------
    gamma: float
        Gamma value.
    eps_init: float
        Initial value for epsilon.
    n_iters: int
        Number of iterations per step.
    """

    def __init__(
        self,
        gamma: float = 1,
        epsilon: float = 1e-4,
        n_iters: int = 10,
        loss: str = "logistic",
    ):
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.loss = loss

    def _loss(self, X, y, task_idxs):
        num_tasks = len(task_idxs)
        W = np.zeros((X.shape[1], num_tasks))
        cost = 0
        err = 0
        reg = 0
        for t_id in range(num_tasks):
            task_x = X[task_idxs[t_id]]
            task_y = y[task_idxs[t_id]]

            if self.loss == "logistic":
                model = liblinearutil.train(
                    task_y, task_x, f"-s 0 -e 1e-3 -c {self.gamma} -q"
                )
                w, b = model.get_decfun()
            elif self.loss == "svm":
                model = liblinearutil.train(
                    task_y, task_x, f"-s 2 -e 1e-3 -c {self.gamma} -q"
                )
                w, b = model.get_decfun()
            elif self.loss == "ols":
                K = task_x.T @ task_x
                mat_inv = np.linalg.inv(K + self.gamma * np.eye(K.shape[0]))
                w = mat_inv @ task_x.T @ task_y

                # K = task_x @ task_x.T  # Linear kernel
                # mat_inv = np.linalg.inv(K + self.gamma * np.eye(K.shape[0]))
                # a = mat_inv @ task_x.T @ task_y
                # cost += self.gamma * task_y @ a
                # err += self.gamma**2 * a.T @ a
                # W[:, t_id] = task_x.T @ a
            w = np.asarray(w)
            W[:, t_id] = w
            _cost = self.gamma * w.T @ w
            cost += _cost
        reg = cost - err
        return W, cost, err, reg

    def Dmin(self, a):
        return a / a.sum()

    def Dmin_eps(self, a, eps):
        return self.Dmin(np.sqrt(a**2 + eps))

    def fit(self, X, y, tasks):
        self._fit_feat(X, y, tasks)
        return self

    def _fit_feat(self, X, y, tasks):
        task_ids = np.unique(tasks)
        task_idxs = [np.flatnonzero(tasks == t) for t in task_ids]
        dim = X.shape[1]
        Dini = np.eye(dim) / dim

        def fmeth(a):
            _a = np.zeros_like(a)
            _a[a > _EPS] = 1 / a[a > _EPS]
            return _a

        # train_alternating
        U, s, _ = np.linalg.svd(Dini, hermitian=True)
        fS = np.sqrt(fmeth(s))
        fS[fS > _EPS] = 1 / fS[fS > _EPS]
        fD_isqrt = U @ np.diag(fS) @ U.T
        _costs = np.empty((self.n_iters, 3))
        for i in range(self.n_iters):
            W, cost, err, reg = self._loss(X @ fD_isqrt, y, task_idxs)
            W = fD_isqrt @ W

            _costs[i] = [cost, err, reg]

            U, s, _ = np.linalg.svd(W)
            if dim > len(task_ids):
                s = np.r_[s, np.zeros(dim - len(task_ids))]
            Smin = self.Dmin_eps(s, eps=self.epsilon)
            D = U @ np.diag(Smin) @ U.T

            U, s, _ = np.linalg.svd(D, hermitian=True)
            fS = np.sqrt(fmeth(s))
            fS[fS > _EPS] = 1 / fS[fS > _EPS]
            fD_isqrt = U @ np.diag(fS) @ U.T
        # train_alternating

        s = np.linalg.svd(D, compute_uv=False)
        _costs[:, [0, 2]] = (
            _costs[:, [0, 2]] + self.gamma * self.epsilon * fmeth(s).sum()
        )
        self.mineps_ = self.epsilon
        self.W_ = W
        self.D_ = D
        self.costs_ = [_costs]

    def predict(self, X):
        return X @ self.W_


class MTFLMulticlass(BaseEstimator, ClassifierMixin):
    """Linear regressor based on MTFL. Code adapted from
    https://github.com/argyriou/multi_task_learning

    Parameters
    ----------
    gamma: float
        Gamma value.
    eps_init: float
        Initial value for epsilon.
    n_iters: int
        Number of iterations per step.
    """

    def __init__(
        self,
        gamma: float = 1,
        epsilon: float = 1e-4,
        n_iters: int = 10,
        loss: str = "logistic",
    ):
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.loss = loss

    def _loss(self, X, y, task_idxs):
        num_tasks = len(task_idxs)
        classes = np.unique(y)
        W = np.zeros((len(classes), X.shape[1], num_tasks))
        cost = np.zeros(len(classes))
        err = np.zeros_like(cost)
        for t_id in range(num_tasks):
            task_x = X[task_idxs[t_id]]
            task_y = y[task_idxs[t_id]]

            task_classes = np.unique(task_y)
            if self.loss == "logistic":
                model = liblinearutil.train(
                    task_y, task_x, f"-s 0 -e 1e-3 -c {self.gamma} -q"
                )
                for i, c in enumerate(task_classes):
                    W[c, :, t_id] = model.get_decfun(label_idx=i)[0]
            elif self.loss == "svm":
                model = liblinearutil.train(
                    task_y, task_x, f"-s 2 -e 1e-3 -c {self.gamma} -q"
                )
                for i, c in enumerate(task_classes):
                    W[c, :, t_id] = model.get_decfun(label_idx=i)[0]
            _cost = self.gamma * (W[:, :, t_id] ** 2).sum(-1)
            cost += _cost
        reg = cost - err
        return W, cost, err, reg

    def Dmin(self, a):
        return a / a.sum()

    def Dmin_eps(self, a, eps):
        return self.Dmin(np.sqrt(a**2 + eps))

    def fit(self, X, y, tasks):
        self.classes_, y = np.unique(y, return_inverse=True)
        self._fit_feat(X, y, tasks)
        return self

    def _fit_feat(self, X, y, tasks):
        task_ids = np.unique(tasks)
        task_idxs = [np.flatnonzero(tasks == t) for t in task_ids]
        n_classes = len(self.classes_)
        dim = X.shape[1]
        Dini = np.tile(np.eye(dim) / dim, (n_classes, 1, 1))

        def fmeth(a):
            _a = np.zeros_like(a)
            _a[a > _EPS] = 1 / a[a > _EPS]
            return _a

        # train_alternating
        U, s, _ = np.linalg.svd(Dini, hermitian=True)
        fS = np.sqrt(fmeth(s))
        fS[fS > _EPS] = 1 / fS[fS > _EPS]
        fD_isqrt = U @ np.apply_along_axis(np.diag, 1, fS) @ U.transpose(0, 2, 1)

        _costs = np.empty((self.n_iters, n_classes, 3))
        for i in range(self.n_iters):
            W, cost, err, reg = self._loss((X @ fD_isqrt).mean(0), y, task_idxs)
            W = fD_isqrt @ W

            _costs[i] = np.c_[cost, err, reg]

            U, s, _ = np.linalg.svd(W)
            if dim > len(task_ids):
                s = np.c_[
                    s,
                    np.zeros(
                        (
                            n_classes,
                            dim - len(task_ids),
                        )
                    ),
                ]
            Smin = self.Dmin_eps(s, eps=self.epsilon)
            D = U @ np.apply_along_axis(np.diag, 1, Smin) @ U.transpose(0, 2, 1)

            U, s, _ = np.linalg.svd(D, hermitian=True)
            fS = np.sqrt(fmeth(s))
            fS[fS > _EPS] = 1 / fS[fS > _EPS]
            fD_isqrt = U @ np.apply_along_axis(np.diag, 1, fS) @ U.transpose(0, 2, 1)
        # train_alternating

        s = np.linalg.svd(D, compute_uv=False)
        _costs[:, [0, 2]] = (
            _costs[:, [0, 2]] + self.gamma * self.epsilon * fmeth(s).sum()
        )
        self.mineps_ = self.epsilon
        self.W_ = W
        self.D_ = D
        self.costs_ = [_costs]

    def decision_function(self, X):
        return (X @ self.W_).mean(-1)

    def predict(self, X):
        return self.classes_[(X @ self.W_).mean(-1).argmax(0)]


class MTFLClassifier(BaseBinaryMTFL, ClassifierMixin):
    """Classifier based on MTFL. Code adapted from
    https://github.com/argyriou/multi_task_learning

    Parameters
    ----------
    gamma: float
        Gamma value.
    eps_init: float
        Initial value for epsilon.
    n_iters: int
        Number of iterations per step.
    loss: str
        Loss to use.
    aggregate: str
        Method of aggregation. One of "logistic", "weighted",
        "unweighted".
    """

    def __init__(
        self,
        gamma: float = 1,
        epsilon: float = 1e-4,
        n_iters: int = 10,
        loss: str = "svm",
        aggregate: str = "weighted",
    ):
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.loss = loss
        self.aggregate = aggregate

    def fit(self, X, y, tasks):
        if len(np.unique(y)) > 2:
            raise ValueError("`y` should have only two classes")
        y = y.copy()
        y[y == 0] = -1
        self._fit_feat(X, y, tasks)
        if self.aggregate == "logistic":
            self.logistic_ = LogisticRegression(penalty="none")
            self.logistic_.fit(super().predict(X), y)
        return self

    def decision_function(self, X):
        preds = super().predict(X)
        if self.aggregate == "logistic":
            return self.logistic_.decision_function(preds)
        return preds.mean(-1)

    def predict(self, X):
        preds = super().predict(X)
        if self.aggregate == "logistic":
            preds = self.logistic_.predict(preds)
        elif self.aggregate == "weighted":
            preds = np.sign(preds.mean(-1))
        elif self.aggregate == "unweighted":
            preds = np.sign(np.sign(preds).mean(-1))
        preds[preds == -1] = 0
        return preds



"""Scikit-learn implementations of various models"""

from typing import Callable

from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier

from .svm import PrecomputedSVC

CLASSIFIER_MAP = {"svm": PrecomputedSVC, "rf": RandomForestClassifier}


def get_sk_model_fn(name: str) -> Callable[..., BaseEstimator]:
    """Get a scikit-learn model class.

    Parameters
    ----------
    name: str
        The model name.

    Returns
    -------
    model_fn: callable
        A method that takes arguments and returns a Model instance.
    """
    return CLASSIFIER_MAP[name]


def get_sk_model(name: str, **kwargs) -> BaseEstimator:
    """Get a scikit-learn model by name.

    Parameters
    ----------
    name: str
        The model name.

    Returns
    -------
    model: BaseEstimator
        The unfit scikit-learn classifier.
    """
    return get_sk_model_fn(name)(**kwargs)



from itertools import chain, combinations, permutations, zip_longest
from typing import (
    Any,
    Callable,
    Container,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    overload,
)

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")


def itmap(s: Callable[[T1], T2]):
    """Returns a new map function that additionally maps tuples to
    tuples and lists to lists.

    Parameters
    ----------
    s: callable
        A callable that converts objects of type T to type S.

    Returns
    -------
    callable
        A function that converts objects of type T to type S, lists of T
        to lists of S, and tuples of T to tuples of S.
    """

    @overload
    def _map(x: T1) -> T2:
        ...

    @overload
    def _map(x: List[T1]) -> List[T2]:
        ...

    @overload
    def _map(x: Tuple[T1, ...]) -> Tuple[T2, ...]:
        ...

    def _map(x):
        if isinstance(x, (list, tuple)):
            return type(x)(s(y) for y in x)
        else:
            return s(x)

    return _map


def ordered_intersect(a: Iterable, b: Container) -> List:
    """Returns a list of the intersection of `a` and `b`, in the order
    elements appear in `a`.

    Parameters
    ----------
    a: iterable
        Iterable with the order of elements.
    b: container
        Container to test inclusion of elements in `a`.

    Returns
    -------
    list
        The resulting elements, in order.
    """
    return [x for x in a if x in b]


def filter_kwargs(kwargs: Dict[str, Any], method: Callable) -> Dict[str, Any]:
    """Removes incompatible keyword arguments. This ignores any
    `**kwargs` catchall in method signature, and only returns args
    specifically present as keyhwords in the method signature which are
    also not positional only.

    Parameters
    ----------
    params: dict
        Keyword arguments to pass to method.
    method: callable
        The method for which to check valid parameters.

    Returns
    -------
    dict
        Filtered keyword arguments.
    """
    import inspect

    meth_params = inspect.signature(method).parameters
    kwargs = kwargs.copy()
    for key in set(kwargs.keys()):
        if (
            key not in meth_params
            or meth_params[key].kind == inspect.Parameter.POSITIONAL_ONLY
        ):
            del kwargs[key]
    return kwargs


def batch_iterable(
    it: Iterable[T], batch_size: int, fillvalue: Any = None
) -> Iterable[Tuple[T, ...]]:
    """Batches an iterable into chunks of size `batch_size`.

    Parameters
    ----------
    it: iterable
        The iterable to batch.
    batch_size: int
        The size of each batch/chunk.
    fillvalue:
        An optional fill value if the final batch isn't full (i.e.
        `batch_size` doesn't divide the iterable length.)

    Yields
    -------
    tuple
        A tuple of length `batch_size` with successive elements from the
        original iterable.
    """
    # From the itertools recipes, to chunk an iterator
    yield from zip_longest(*[iter(it)] * batch_size, fillvalue=fillvalue)


def subsets(it: Iterable[T], max_size: Optional[int] = None) -> Iterable[Tuple[T, ...]]:
    """Iterate over all subsets of the iterable `it`, up to a given
    maximum size. This will generate subsets in size order and then
    index-sorted order (i.e. the order items appear in `it`).

    Parameters
    ----------
    it: iterable
        The iterable from which to generate subsets.
    max_size: int, optional
        The maximum size of generated subsets. If not given, the size of
        `it` is determined by creating a list of `it`'s elements.

    Yields
    ------
    tuple
        The next generated subset.
    """
    if max_size is None:
        it = list(it)
        max_size = len(it)
    yield from chain(*(combinations(it, i) for i in range(max_size + 1)))


def ordered_subsets(
    it: Iterable[T], max_size: Optional[int] = None
) -> Iterable[Tuple[T, ...]]:
    """Iterate over all ordered subsets of the iterable `it`, up to a
    given maximum size. This will generate subsets in size order and
    then index-sorted order (i.e. the order items appear in `it`).

    Parameters
    ----------
    it: iterable
        The iterable from which to generate ordered subsets.
    max_size: int, optional
        The maximum size of generated subsets. If not given, the size of
        `it` is determined by creating a list of `it`'s elements.

    Yields
    ------
    tuple
        The next generated ordered subset.
    """
    if max_size is None:
        it = list(it)
        max_size = len(it)
    yield from chain(*(permutations(it, i) for i in range(max_size + 1)))



from pathlib import Path
from typing import Union

import click
import matplotlib.pyplot as plt

from ertk.dataset import read_features


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("instance", type=str, default="2")
def main(input: Path, instance: str):
    """Displays plot of INSTANCE in INPUT. INSTANCE can either be a
    numeric index, a range of indices using numpy slice notation or a
    named instance.
    """

    data = read_features(input)
    if instance.isdigit():
        idx: Union[int, slice] = int(instance)
    else:
        _i = instance.find(":")
        if _i != -1:
            start = int(instance[:_i])
            end = int(instance[_i + 1 :])
            idx = slice(start, end)
        else:
            idx = data.names.index(instance)
    arr = data.features[idx]
    names = data.names[idx]
    print(names)

    plt.figure()
    plt.imshow(arr, aspect="equal", origin="upper", interpolation="nearest")
    plt.xlabel("Features")
    plt.ylabel("Instance" if len(names) > 1 else "Time")
    plt.show()


if __name__ == "__main__":
    main()



"""Implementation of the model from [1].

[1] S. Latif, R. Rana, S. Khalifa, R. Jurdak, and J. Epps, 'Direct
Modelling of Speech Emotion from Raw Speech', arXiv:1904.03833 [cs,
eess], Jul. 2019, Accessed: Feb. 10, 2020. [Online]. Available:
http://arxiv.org/abs/1904.03833.
"""

import tensorflow as tf
from tensorflow.keras.layers import (
    LSTM,
    BatchNormalization,
    Conv1D,
    Conv2D,
    Dense,
    Dropout,
    Input,
    MaxPool1D,
    MaxPool2D,
    ReLU,
    Reshape,
    concatenate,
)
from tensorflow.keras.models import Model

__all__ = ["model"]


def model(n_classes: int, n_features: int = 1):
    if n_features != 1:
        raise ValueError("This model can only accept waveform input.")

    inputs = Input((None, 1), name="input")

    # 3x 1D convs and maxpool over time
    conv_1 = Conv1D(
        40,
        400,
        strides=160,
        padding="same",
        kernel_initializer="he_normal",
        name="conv_25ms",
    )(inputs)
    conv_1 = BatchNormalization()(conv_1)
    conv_1 = ReLU()(conv_1)
    maxpool1 = MaxPool1D(2, name="maxpool_1")(conv_1)

    conv_2 = Conv1D(
        40,
        240,
        strides=160,
        padding="same",
        kernel_initializer="he_normal",
        name="conv_15ms",
    )(inputs)
    conv_2 = BatchNormalization()(conv_2)
    conv_2 = ReLU()(conv_2)
    maxpool2 = MaxPool1D(2, name="maxpool_2")(conv_2)

    conv_3 = Conv1D(
        40,
        1600,
        strides=160,
        padding="same",
        kernel_initializer="he_normal",
        name="conv_100ms",
    )(inputs)
    conv_3 = BatchNormalization()(conv_3)
    conv_3 = ReLU()(conv_3)
    maxpool3 = MaxPool1D(2, name="maxpool_3")(conv_3)

    concat = concatenate([maxpool1, maxpool2, maxpool3], axis=-1)
    concat = tf.expand_dims(concat, axis=-1)

    # 2D conv and maxpool over features and time
    conv2d = Conv2D(
        32, (2, 2), name="conv2d", padding="same", kernel_initializer="he_normal"
    )(concat)
    conv2d = BatchNormalization()(conv2d)
    conv2d = ReLU()(conv2d)
    maxpool2d = MaxPool2D((2, 2), name="maxpool2d", padding="same")(conv2d)

    # LSTM
    flattened = Reshape((-1, 1920), name="flatten")(maxpool2d)
    lstm = LSTM(128)(flattened)
    dropout = Dropout(0.3)(lstm)

    dense = Dense(
        1024,
        activation="relu",
        kernel_initializer="he_normal",
    )(dropout)
    x = Dense(
        n_classes,
        activation="softmax",
        kernel_initializer="he_normal",
    )(dense)
    return Model(inputs=inputs, outputs=x)



from .array import (
    batch_arrays_by_length,
    clip_arrays,
    flat_to_inst,
    frame_array,
    frame_arrays,
    inst_to_flat,
    is_mono_audio,
    make_array_array,
    pad_arrays,
    shuffle_multiple,
    transpose_time,
)
from .generic import (
    batch_iterable,
    filter_kwargs,
    itmap,
    ordered_intersect,
    ordered_subsets,
    subsets,
)
from .misc import TqdmParallel
from .types import PathOrStr, ScoreFunction



from pathlib import Path

import click
import numpy as np

from ertk.config import get_arg_mapping
from ertk.tensorflow import get_tf_model
from ertk.tensorflow.utils import init_gpu_memory_growth, tf_dataset_gen, tf_dataset_mem


@click.command()
@click.argument("model_name")
@click.option("--batch_size", type=int, default=32, help="Batch size.")
@click.option(
    "--summary_only", is_flag=True, help="Print summary but don't train on dummy data."
)
@click.option(
    "--memory/--generated",
    default=False,
    help="Have all data in memory or generate batches.",
)
@click.option("--model_args", type=click.Path(exists=True, path_type=Path))
@click.option("--train", type=int, default=8000, help="Number of training instances.")
@click.option("--valid", type=int, default=800, help="Number of validation instances.")
@click.option("--features", type=int, default=1024, help="Dimensionality of input.")
@click.option("--steps", type=int, default=512, help="Length of sequence.")
def test_fit(
    model_name: str,
    batch_size: int,
    summary_only: bool,
    memory: bool,
    model_args: Path,
    train: int,
    valid: int,
    features: int,
    steps: int,
):
    """Creates a TensorFlow model and fits it with random data. This is
    useful for seeing how much memory will be used with different batch
    sizes and how long training takes.
    """

    init_gpu_memory_growth()

    args = get_arg_mapping(model_args) if model_args else {}
    model = get_tf_model(model_name, n_features=features, n_classes=4, **args)
    model.summary()

    if summary_only:
        return

    shape = (features,) if len(model.input_shape) == 2 else (steps, features)

    rng = np.random.default_rng()
    train_x = rng.random((train,) + shape, dtype=np.float32)
    train_y = rng.integers(4, size=train)
    valid_x = rng.random((valid,) + shape, dtype=np.float32)
    valid_y = rng.integers(4, size=valid)

    if memory:
        train_data = tf_dataset_mem(
            train_x, train_y, batch_size=batch_size, shuffle=True
        )
        valid_data = tf_dataset_mem(
            valid_x, valid_y, batch_size=batch_size, shuffle=False
        )
    else:
        train_data = tf_dataset_gen(
            train_x, train_y, batch_size=batch_size, shuffle=True
        )
        valid_data = tf_dataset_gen(
            valid_x, valid_y, batch_size=batch_size, shuffle=False
        )

    model.compile(loss="sparse_categorical_crossentropy")
    model.fit(train_data, epochs=2, validation_data=valid_data, verbose=True)


if __name__ == "__main__":
    test_fit()



import json
import warnings
from functools import partial
from pathlib import Path
from typing import Tuple

import click
import numpy as np
from click_option_group import optgroup
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from ertk.classification import get_balanced_sample_weights, train_val_test
from ertk.config import get_arg_mapping
from ertk.dataset import load_multiple
from ertk.train import ValidationSplit
from ertk.transform import SequenceTransformWrapper


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option("--features", required=True, help="Features to load.")
@click.option("--clf", "clf_type", required=True, help="Classifier to use.")
@click.option("--train", required=True, help="Train data.")
@click.option("--valid", required=True, help="Validation data.")
@click.option("--test", required=True, help="Test data.")
@optgroup.group("Results options")
@optgroup.option("--results", type=Path, help="Results directory.")
@optgroup.option("--logdir", type=Path, help="TF/PyTorch logs directory.")
@optgroup.group("Data options")
@optgroup.option("--label", default="label", help="Label annotation to use.")
@optgroup.option(
    "--subset", multiple=True, default=["default"], help="Subset selection."
)
@optgroup.option("--map_groups", multiple=True, help="Group mapping.")
@optgroup.option(
    "--sel_groups",
    multiple=True,
    help="Group selection. This is a map from partition to group.",
)
@optgroup.option(
    "--clip_seq", type=int, help="Clip sequences to this length (before pad)."
)
@optgroup.option(
    "--pad_seq", type=int, help="Pad sequences to multiple of this length (after clip)."
)
@optgroup.group("Experiment options")
@optgroup.option(
    "--n_gpus", type=int, default=1, show_default=True, help="Number of GPUs to use."
)
@optgroup.option(
    "--reps",
    type=int,
    default=1,
    show_default=True,
    help="The number of repetitions to do per test.",
)
@optgroup.option(
    "--normalise",
    default="online",
    show_default=True,
    help="Normalisation method. 'online' means use training data for normalisation.",
)
@optgroup.option(
    "--transform",
    type=click.Choice(["std", "minmax"]),
    default="std",
    show_default=True,
    help="Transformation class.",
)
@optgroup.option(
    "--balanced/--imbalanced", default=True, help="Balances class weights."
)
@optgroup.group("Misc. options")
@optgroup.option(
    "--verbose",
    type=int,
    default=0,
    help="Verbosity. -1=nothing, 0=dataset+results, 1=INFO, 2=DEBUG",
)
@optgroup.group("Model-specific options")
@optgroup.option(
    "--clf_args",
    "clf_args_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    multiple=True,
    help="File containing keyword arguments to give to model initialisation.",
)
@optgroup.option(
    "--param_grid",
    "param_grid_file",
    type=click.Path(exists=True, path_type=Path),
    multiple=True,
    help="File with parameter grid data.",
)
@optgroup.option("--learning_rate", type=float, default=1e-4, show_default=True)
@optgroup.option("--batch_size", type=int, default=64, show_default=True)
@optgroup.option("--epochs", type=int, default=50, show_default=True)
def main(
    clf_type: str,
    input: Tuple[Path],
    features: str,
    train: str,
    valid: str,
    test: str,
    label: str,
    param_grid_file: Tuple[Path],
    results: Path,
    logdir: Path,
    reps: int,
    normalise: str,
    transform: str,
    balanced: bool,
    subset: Tuple[str],
    map_groups: Tuple[str],
    sel_groups: Tuple[str],
    clip_seq: int,
    pad_seq: int,
    verbose: int,
    clf_args_file: Tuple[Path],
    learning_rate: float,
    batch_size: int,
    epochs: int,
    n_gpus: int,
):
    """Runs training on the given INPUT dataset, using training set.
    Validation is done on validation set and results reported on test
    set. Metrics are optionally written to a results file.
    """

    np.set_printoptions(precision=4, threshold=10)
    if verbose > 0:
        import logging

        logging.basicConfig(level=logging.DEBUG if verbose > 1 else logging.INFO)
        np.set_printoptions()

    if len(subset) == 1:
        dataset = load_multiple(input, features, subsets=subset[0])
    else:
        subset_map = {}
        for m in subset:
            subset_map.update(get_arg_mapping(m))
        dataset = load_multiple(input, features, subsets=subset_map)

    for m in map_groups:
        mapping = get_arg_mapping(m)
        for part in mapping:
            if part not in dataset.partitions:
                warnings.warn(
                    f"Partition {part} cannot be mapped as it is not in the dataset."
                )
                continue
            dataset.map_groups(part, mapping[part])
    for m in sel_groups:
        mapping = get_arg_mapping(m)
        for part in mapping:
            if part not in dataset.partitions:
                warnings.warn(
                    f"Partition {part} cannot be selected as it is not in the dataset."
                )
                continue
            keep = mapping[part]
            if isinstance(keep, str):
                keep = [keep]
            dataset.remove_groups(part, keep=keep)

    if clip_seq:
        dataset.clip_arrays(clip_seq)
    if pad_seq:
        dataset.pad_arrays(pad_seq)

    transformer = {"std": StandardScaler, "minmax": MinMaxScaler}[transform]()
    if normalise == "none":
        transformer = None
    elif normalise == "online" and len(dataset.x[0].shape) > 1:
        transformer = SequenceTransformWrapper(transformer, "feature")
    elif normalise != "online":
        dataset.normalise(normaliser=transformer, partition=normalise)
        transformer = None

    sample_weight = None
    if balanced:
        sample_weight = get_balanced_sample_weights(dataset.get_annotations(label))
    fit_params = {"sample_weight": sample_weight}

    model_args = {}
    for file in clf_args_file:
        model_args.update(get_arg_mapping(file))
    param_grid = {}
    for file in param_grid_file:
        param_grid.update(get_arg_mapping(file))

    train_indices = dataset.get_idx_for_split(train)
    valid_indices = dataset.get_idx_for_split(valid)
    test_indices = dataset.get_idx_for_split(test)

    clf_lib, clf_type = clf_type.split("/", maxsplit=1)
    if clf_lib == "sk":
        from sklearn.base import clone
        from sklearn.model_selection import GridSearchCV

        from ertk.sklearn.models import get_sk_model

        class GridSearchWrapper(GridSearchCV):
            """Simple hack to avoid retraining on train + val data."""

            def fit(self, X, y=None, *, groups=None, **fit_params):
                self.refit = False
                super().fit(X, y=y, groups=groups, **fit_params)
                self.best_estimator_ = clone(
                    clone(self.estimator).set_params(**self.best_params_)
                )
                train, _ = next(iter(self.cv.split(X, y, groups)))
                self.best_estimator_.fit(X[train], y[train], **fit_params)
                return self.best_estimator_

        clf = get_sk_model(clf_type, **model_args)
        clf = GridSearchWrapper(
            Pipeline([("transform", transformer), ("clf", clf)]),
            {f"clf__{k}": v for k, v in param_grid.items()},
            scoring="balanced_accuracy",
            cv=ValidationSplit(
                np.arange(len(train_indices), len(train_indices) + len(valid_indices))
            ),
            verbose=max(verbose, 0),
            n_jobs=-1,
            refit=False,
        )
        params = param_grid
        # A hack so that the ValidationSplit works properly.
        train_indices = np.concatenate([train_indices, valid_indices])
    elif clf_lib == "tf":
        from tensorflow.keras.callbacks import TensorBoard
        from tensorflow.keras.optimizers import Adam

        from ertk.tensorflow import get_tf_model
        from ertk.tensorflow.classification import tf_classification_metrics

        callbacks = []
        if logdir is not None:
            callbacks.append(
                TensorBoard(
                    log_dir=logdir,
                    profile_batch=0,
                    write_graph=False,
                    write_images=False,
                )
            )
        fit_params.update(
            {
                "epochs": epochs,
                "callbacks": callbacks,
                "batch_size": batch_size,
                "data_fn": None,
                "n_gpus": n_gpus,
            }
        )
        model_args.update(
            {"n_features": dataset.n_features, "n_classes": dataset.n_classes}
        )

        def model_fn():
            model = get_tf_model(clf_type, **model_args)
            model.compile(
                Adam(learning_rate=learning_rate),
                loss="sparse_categorical_crossentropy",
                metrics=tf_classification_metrics(),
            )
            return Pipeline([("transform", transformer), ("clf", model)])

        params = {
            "optimiser": "adam",
            "learning_rate": learning_rate,
            "batch_size": batch_size,
            **model_args,
            **fit_params,
        }
        clf = model_fn
    elif clf_lib == "pt":
        import torch.optim

        from ertk.pytorch import get_pt_model

        optim_fn = partial(torch.optim.Adam, lr=learning_rate)
        fit_params.update(
            {
                "log_dir": logdir,
                "max_epochs": epochs,
                "batch_size": batch_size,
                "optim_fn": optim_fn,
                "n_gpus": n_gpus,
            }
        )
        model_args.update(
            {"n_features": dataset.n_features, "n_classes": dataset.n_classes}
        )

        def model_fn():
            model = get_pt_model(clf_type, **model_args)
            return model if transformer is None else (transformer, model)

        params = {
            "optimiser": "adam",
            "learning_rate": learning_rate,
            **model_args,
            **fit_params,
        }
        clf = model_fn
    else:
        raise ValueError(f"Invalid classifier type {clf_type}")

    if verbose > -1:
        print("== Dataset ==")
        print(dataset)
        print("== Train/valid/test settings ==")
        print(f"Using {dataset.n_classes}-class classification.")
        print(f"Using classifier {clf}.")
        print(f"Using fit_params={fit_params}")
        print(f"Training set: {train}")
        print(f"Validation set: {valid}")
        print(f"Test set: {test}")
        if balanced:
            print("Using balanced sample weights.")
        if normalise == "online":
            print("Using 'online' normalisation.")
        elif normalise != "none":
            print(f"Normalising globally using {normalise} partition.")

    df = train_val_test(
        clf,
        dataset=dataset,
        train_idx=train_indices,
        valid_idx=valid_indices,
        test_idx=test_indices,
        label=label,
        clf_lib=clf_lib,
        sample_weight=sample_weight,
        verbose=verbose,
        fit_params=fit_params,
    )
    df["clf"] = f"{clf_lib}/{clf_type}"
    df["features"] = features
    df["corpus"] = dataset.corpus
    df["params"] = json.dumps(params, default=str)
    if results:
        results.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(results)
        print(f"Wrote CSV to {results}")
    else:
        print(df)
    if verbose > -1:
        print(df.mean(numeric_only=True).to_string())


if __name__ == "__main__":
    main()



from pathlib import Path

import click
import numpy as np
import pytorch_lightning as pl
import torch
from torch import nn
from torch.optim import Adam
from torch.utils.data import DataLoader, TensorDataset

from ertk.config import get_arg_mapping
from ertk.pytorch import get_pt_model
from ertk.pytorch.utils import LightningWrapper


@click.command()
@click.argument("model_name")
@click.option("--batch_size", type=int, default=32, help="Batch size.")
@click.option("--1D/--2D", "is_1d", default=True, help="Input type.")
@click.option(
    "--summary_only", is_flag=True, help="Print summary but don't train on dummy data."
)
@click.option(
    "--memory/--generated",
    default=False,
    help="Have all data in memory or generate batches.",
)
@click.option("--model_args", type=click.Path(exists=True, path_type=Path))
@click.option("--train", type=int, default=8000, help="Number of training instances.")
@click.option("--valid", type=int, default=800, help="Number of validation instances.")
@click.option("--features", type=int, default=1024, help="Dimensionality of input.")
@click.option("--steps", type=int, default=512, help="Length of sequence.")
def test_fit(
    model_name: str,
    batch_size: int,
    is_1d: bool,
    summary_only: bool,
    memory: bool,
    model_args: Path,
    train: int,
    valid: int,
    features: int,
    steps: int,
):
    """Trains a model for 2 epochs with random data and prints timing
    information for testing purposes. The default number of instances is
    8000, which is larger than the current largest dataset, MSP-IMPROV.
    """

    args = get_arg_mapping(model_args) if model_args else {}
    model = get_pt_model(model_name, n_features=features, n_classes=4, **args)
    if not isinstance(model, pl.LightningModule):
        model = LightningWrapper(model, nn.CrossEntropyLoss(), lambda x: Adam(x, 0.001))
    print(model)
    trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    total_params = sum(p.numel() for p in model.parameters())
    print("Trainable parameters: {}".format(trainable_params))
    print("Total parameters: {}".format(total_params))
    print()

    if summary_only:
        return

    shape = (features,) if is_1d else (steps, features)

    rng = np.random.default_rng()
    train_x = torch.as_tensor(rng.random((train,) + shape, dtype=np.float32))
    train_y = torch.as_tensor(rng.integers(4, size=train))
    valid_x = torch.as_tensor(rng.random((valid,) + shape, dtype=np.float32))
    valid_y = torch.as_tensor(rng.integers(4, size=valid))

    train_data = TensorDataset(train_x, train_y)
    valid_data = TensorDataset(valid_x, valid_y)
    train_dl = DataLoader(
        train_data, batch_size=batch_size, shuffle=True, pin_memory=True
    )
    valid_dl = DataLoader(
        valid_data, batch_size=batch_size, shuffle=False, pin_memory=True
    )

    trainer = pl.Trainer(
        gpus=1,
        logger=False,
        enable_progress_bar=True,
        enable_model_summary=False,
        enable_checkpointing=False,
        max_epochs=2,
    )
    trainer.fit(model, train_dataloaders=train_dl, val_dataloaders=valid_dl)
    preds = torch.cat(trainer.predict(model, valid_dl)).cpu().numpy()
    print(preds)


if __name__ == "__main__":
    test_fit()



from pathlib import Path
from typing import Tuple

import click

from ertk.dataset import Dataset


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option("--subset", default="default")
def main(input: Tuple[Path], subset: str):
    """Print info about datasets in INPUT."""
    for file in input:
        dataset = Dataset(file, subset=subset)
        print(dataset)


if __name__ == "__main__":
    main()



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from ertk.utils import flat_to_inst, inst_to_flat


def group_transform(
    x: np.ndarray,
    groups: np.ndarray,
    transform: TransformerMixin,
    *,
    inplace: bool = False,
    **fit_params,
):
    """Per-group (offline) transformation (e.g. standardisation).

    Parameters
    ----------
    x: np.ndarray
        The data matrix to transform. Each x[i] must be an instance.
    groups: np.ndarray
        Groups assignment for each instance. It must be the case that
        len(groups) == len(x).
    transform:
        The transformation to apply. Must implement fit_transform().
    inplace: bool
        Whether to modify x in-place. Default is False so that a copy is
        made.
    **fit_params:
        Other keyword arguments to pass to the transform.fit() method.

    Returns
    -------
    x: np.ndarray
        The modified data matrix with transformations applied to each
        group individually.
    """
    if not inplace:
        x = x.copy()
    unique_groups = np.unique(groups)
    for g_id in unique_groups:
        flat, slices = inst_to_flat(x[groups == g_id])
        flat = transform.fit_transform(flat, y=None, **fit_params)
        if len(x.shape) == 1 and len(slices) == 1:
            # Special case to avoid issues for vlen arrays
            _arr = np.empty(1, dtype=object)
            _arr[0] = flat
            x[groups == g_id] = _arr
            continue
        x[groups == g_id] = flat_to_inst(flat, slices)
    return x


def instance_transform(
    x: np.ndarray, transform: TransformerMixin, *, inplace: bool = False, **fit_params
):
    """Per-instance transformation (e.g. standardisation).

    Parameters
    ----------
    x: np.ndarray
        The data matrix to transform. Each x[i] must be a 2D instance.
    transform:
        The transformation to apply. Must implement fit_transform().
    inplace: bool
        Whether to modify x in-place. Default is False so that a copy is
        made.
    **fit_params:
        Other keyword arguments to pass to the transform.fit() method.

    Returns
    -------
    x: np.ndarray
        The modified data matrix with transformations applied to each
        instance individually.
    """
    return group_transform(
        x, np.arange(len(x)), transform, inplace=inplace, **fit_params
    )


class GroupTransformWrapper(TransformerMixin, BaseEstimator):
    """Transform that modifies groups independently without storing
    parameters.
    """

    def __init__(self, transformer: TransformerMixin) -> None:
        self.transformer = transformer

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X, groups=None, **fit_params):
        return group_transform(X, groups, self.transformer, inplace=False, **fit_params)


class InstanceTransformWrapper(TransformerMixin, BaseEstimator):
    """Transform that modifies instances independently without storing
    parameters.
    """

    def __init__(self, transformer: TransformerMixin) -> None:
        self.transformer = transformer

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X, **fit_params):
        raise instance_transform(X, self.transformer, inplace=False, **fit_params)


class SequenceTransform(TransformerMixin, BaseEstimator):
    """Transform designed to process sequences of vectors."""

    pass


class SequenceTransformWrapper(SequenceTransform):
    """Wrapper around a scikit-learn transform that can process
    sequences of vectors.

    Parameters
    ----------
    transformer:
        An object which implements the fit_transform() method on a
        collection of 1D vectors.
    method: str
        The method to manipuate the sequence into 1D vectors, one of
        {"feature", "global"}. If "feature" then each feature column of
        the concatenated (2D) input is transformed independently. If
        "global" then the transformer is fitted over the whole input
        including all feature columns.
    """

    def __init__(self, transformer: TransformerMixin, method: str):
        VALID_METHODS = {"feature", "global"}
        self.transformer = transformer
        if method not in VALID_METHODS:
            raise ValueError(f"method '{method}' not in {VALID_METHODS}.")
        self.method = method

    def fit(self, X, y=None, **fit_params):
        flat_x, _ = inst_to_flat(X)
        if self.method == "feature":
            self.transformer.fit(flat_x, y=y, **fit_params)
        elif self.method == "global":
            self.transformer.fit(flat_x.reshape((-1, 1)), y=y, **fit_params)
        return self

    def transform(self, X, **fit_params):
        flat_x, slices = inst_to_flat(X)
        if self.method == "feature":
            flat_x = self.transformer.transform(flat_x, **fit_params)
        elif self.method == "global":
            flat_shape = flat_x.shape
            flat_x = self.transformer.transform(
                flat_x.reshape((-1, 1)), **fit_params
            ).reshape(flat_shape)
        return flat_to_inst(flat_x, slices)



from pathlib import Path
from typing import List, Tuple

import click
import matplotlib.pyplot as plt
import pandas as pd

from ertk.dataset import get_audio_paths


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option("--plot", is_flag=True, help="Plot histogram/bar chart.")
@click.option(
    "--files",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File with names to include for statistics.",
)
def main(input: Tuple[Path], plot: bool, files: Path):
    """Calculate statistics from annotations in INPUT(s). If multiple
    INPUTs are given, additional statistics for the other INPUTs are
    shown for each level of INPUT.
    """

    dfs: List[pd.DataFrame] = []
    for file in input:
        df = pd.read_csv(file, header=0, converters={0: str}).set_index("name")
        if files:
            names = {Path(x).stem for x in get_audio_paths(files)}
            df = df[df.index.isin(names)]
        dfs.append(df)
        col = df[df.columns[0]]
        print(f"== Statistics for {col.name} ==")
        print(col.describe().to_string())
        print()
        if col.dtype == object:
            counts = col.value_counts()
            print("Value counts:")
            print(counts.to_string())
            print()

        if plot:
            fig, ax = plt.subplots()
            ax.set_title(f"Distribution of {file.stem}")
            if col.dtype == object:
                col.value_counts().plot(kind="bar", ax=ax)
            else:
                col.plot(kind="hist", ylabel=None, ax=ax)
            fig.tight_layout()

    if len(input) > 1:
        print("== Pairwise tables ==")
        joined = dfs[0]
        refcol = joined.columns[0]
        for df in dfs[1:]:
            joined = joined.join(df)
        for col in joined.columns[1:]:
            print(f"For column '{col}':")
            if plot:
                fig, ax = plt.subplots()
                ax.set_title(f"Distribution of {col} per {refcol}")

            if joined[refcol].dtype == object:
                if joined[col].dtype == object:
                    table = joined.groupby([refcol, col]).size().unstack(fill_value=0)
                    print(table.to_string())
                    if plot:
                        table.plot(kind="bar", stacked=True, ax=ax)
                    if len(joined.columns) > 2:
                        uniq = (
                            joined.groupby([refcol, col])
                            .nunique()
                            .unstack(fill_value=0)
                        )
                        print("Unique")
                        print(uniq.to_string())
                else:
                    print(joined.groupby([refcol])[col].describe())
                    if plot:
                        table = joined.groupby([refcol])[col].hist(
                            bins=5, alpha=0.3, legend=True, ax=ax
                        )
            else:
                joined.plot(kind="scatter", x=refcol, y=col, ax=ax)
            print()
            if plot:
                fig.tight_layout()

    if plot:
        plt.show()


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F
from torch import nn


class Model(nn.Module):
    def __init__(self, n_features: int, n_classes: int):
        super().__init__()

        self.conv8 = nn.Sequential(
            nn.Conv1d(n_features, 384, kernel_size=8, padding=20), nn.ReLU(inplace=True)
        )
        self.conv16 = nn.Sequential(
            nn.Conv1d(n_features, 384, kernel_size=16, padding=20),
            nn.ReLU(inplace=True),
        )
        self.conv32 = nn.Sequential(
            nn.Conv1d(n_features, 384, kernel_size=32, padding=20),
            nn.ReLU(inplace=True),
        )
        self.conv64 = nn.Sequential(
            nn.Conv1d(n_features, 384, kernel_size=64, padding=20),
            nn.ReLU(inplace=True),
        )
        self.dense = nn.Sequential(
            nn.Linear(1536, 1024),
            nn.ReLU(inplace=True),
            nn.Linear(1024, 1024),
            nn.ReLU(inplace=True),
            nn.Linear(1024, n_classes),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # (batch, steps, features) -> (batch, features, steps)
        x.transpose_(1, 2)
        steps = x.size(-1)
        c1 = F.max_pool1d(self.conv8(x), steps - 7)
        c2 = F.max_pool1d(self.conv16(x), steps - 15)
        c3 = F.max_pool1d(self.conv32(x), steps - 31)
        c4 = F.max_pool1d(self.conv64(x), steps - 63)
        x = torch.cat([c1, c2, c3, c4], -2).squeeze(-1)
        out = self.dense(x)
        return out



"""Implementation of the final full model architecture from [1].

[1] Z. Aldeneh and E. Mower Provost, 'Using regional saliency for speech
emotion recognition', in 2017 IEEE International Conference on
Acoustics, Speech and Signal Processing (ICASSP), Mar. 2017, pp.
2741–2745, doi: 10.1109/ICASSP.2017.7952655.
"""

from tensorflow.keras import Model
from tensorflow.keras.layers import Conv1D, Dense, GlobalMaxPool1D, Input, concatenate

__all__ = ["model"]


def model(n_features: int, n_classes: int) -> Model:
    inputs = Input(shape=(None, n_features), name="input")
    x = Conv1D(384, 8, activation="relu", kernel_initializer="he_normal", name="conv8")(
        inputs
    )
    c1 = GlobalMaxPool1D(name="maxpool_1")(x)

    x = Conv1D(
        384, 16, activation="relu", kernel_initializer="he_normal", name="conv16"
    )(inputs)
    c2 = GlobalMaxPool1D(name="maxpool_2")(x)

    x = Conv1D(
        384, 32, activation="relu", kernel_initializer="he_normal", name="conv32"
    )(inputs)
    c3 = GlobalMaxPool1D(name="maxpool_3")(x)

    x = Conv1D(
        384, 64, activation="relu", kernel_initializer="he_normal", name="conv64"
    )(inputs)
    c4 = GlobalMaxPool1D(name="maxpool_4")(x)

    x = concatenate([c1, c2, c3, c4], name="concatenate")
    x = Dense(1024, activation="relu", kernel_initializer="he_normal", name="dense_1")(
        x
    )
    x = Dense(1024, activation="relu", kernel_initializer="he_normal", name="dense_2")(
        x
    )
    x = Dense(
        n_classes,
        activation="softmax",
        kernel_initializer="he_normal",
        name="emotion_prediction",
    )(x)
    return Model(inputs=inputs, outputs=x)



import click

from .create_cv_dirs import main as create_cv_dirs_main
from .grid_to_conf import main as grid_to_conf_main
from .names_to_filenames import main as names_to_filenames_main
from .parallel_jobs import main as parallel_jobs_main
from .split_chat import main as split_chat_main
from .split_elan import main as split_elan_main


@click.group(no_args_is_help=True)
def cli():
    """ERTK utility scripts."""
    click.echo("ERTK utility CLI")


cli.add_command(create_cv_dirs_main, "create_cv_dirs")
cli.add_command(grid_to_conf_main, "grid_to_conf")
cli.add_command(parallel_jobs_main, "parallel_jobs")
cli.add_command(names_to_filenames_main, "names_to_filenames")
cli.add_command(split_chat_main, "split_chat")
cli.add_command(split_elan_main, "split_elan")



from dataclasses import dataclass
from typing import Callable, Optional

import pytorch_lightning as pl
import torch
import torch.nn as nn


@dataclass
class MTLTaskConfig:
    name: str
    weight: float
    output_dim: int
    loss: nn.Module


class LightningWrapper(pl.LightningModule):
    """Wraps a torch Module instance to make it compatible with PyTorch
    Lightning. Assumes that the Module only outputs a single tensor.

    Parameters
    ----------
    model: nn.Module
        The model to wrap.
    loss: callable
        The loss function. Must take the output and ground truth and
        return a loss tensor.
    optim: torch.optim.Optimizer
        The optimiser to use.
    """

    def __init__(
        self,
        model: nn.Module,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        optim_fn: Callable[..., torch.optim.Optimizer],
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.model = model
        self.loss = loss
        self.optim_fn = optim_fn
        self.save_hyperparameters(
            {"opt": opt.__class__.__name__, "lr": opt.defaults.get("lr", None)}
        )

    def forward(self, x, *args, **kwargs):
        if hasattr(self.model, "preprocess_input"):
            x = self.model.preprocess_input(x)
        return self.model.forward(x, *args, **kwargs)

    def log_loss(
        self, name: str, y: torch.Tensor, yhat: torch.Tensor, sw: Optional[torch.Tensor]
    ):
        loss = self.loss(yhat, y)
        if sw is not None:
            loss = loss.dot(sw) / sw.sum()
        else:
            loss = loss.mean()
        self.log(f"{name}_loss", loss.item())
        return loss

    def training_step(self, batch, batch_idx: int):
        x, y, *sw = batch
        sw = sw[0] if len(sw) > 0 else None
        yhat = self.forward(x)
        loss = self.log_loss("train", y, yhat, sw)
        return loss

    def validation_step(self, batch, batch_idx: int):
        x, y, *sw = batch
        sw = sw[0] if len(sw) > 0 else None
        yhat = self.forward(x)
        self.log_loss("val", y, yhat, sw)

    def predict_step(self, batch, batch_idx: int):
        x, *_ = batch
        return self.forward(x)

    def configure_optimizers(self):
        return self.optim_fn(self.parameters())



import importlib
from functools import partial
from typing import Callable

from torch.nn import Module


def get_pt_model_fn(name: str, **kwargs) -> Callable[..., Module]:
    """Get a PyTorch model creation function.

    Parameters
    ----------
    name: str
        The model name. This should be the name of the module that
        implements the model.
    **kwargs:
        Keyword arguments to provide to the model creation function.

    Returns
    -------
    model_fn: callable
        A method that takes arguments and returns a Module instance.
    """
    try:
        module = importlib.import_module(f".{name}", __package__)
        model_fn = getattr(module, "Model")
    except (ImportError, AttributeError) as e:
        raise ValueError(f"{name} does not correspond to a valid PyTorch model.") from e
    return partial(model_fn, **kwargs)


def get_pt_model(name: str, **kwargs) -> Module:
    """Get a PyTorch model by name.

    Parameters
    ----------
    name: str
        The model name. This should be the name of the module that
        implements the model.
    **kwargs:
        Keyword arguments to provide to the model creation function.

    Returns
    -------
    model: Module
        The PyTorch model.
    """
    return get_pt_model_fn(name, **kwargs)()



from typing import Dict, Mapping, Optional, Type, TypeVar

import pandas as pd

from ertk.utils import PathOrStr

AT = TypeVar("AT", str, int, float)


def read_annotations(
    filename: PathOrStr, dtype: Optional[Type[AT]] = None
) -> Dict[str, AT]:
    """Returns a dict of the form {name: annotation}."""
    # Need index_col to be False or None due to
    # https://github.com/pandas-dev/pandas/issues/9435
    df = pd.read_csv(filename, index_col=False, header=0, converters={0: str, 1: dtype})
    annotations = df.set_index(df.columns[0])[df.columns[1]].to_dict()
    return annotations


def write_annotations(
    annotations: Mapping[str, object], name: str, path: Optional[PathOrStr] = None
):
    """Write sorted annotations CSV.

    Parameters
    ----------
    annotations: mapping
        A mapping of the form {name: annotation}.
    name: str
        Name of the annotation.
    path: pathlike or str, optional
        Path to write CSV. If None, filename is name.csv
    """
    df = pd.DataFrame.from_dict(annotations, orient="index", columns=[name])
    df.index.name = "name"
    path = path or f"{name}.csv"
    df.sort_index().to_csv(path, header=True, index=True)



# Copyright 2019 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Feature computation for YAMNet."""

import numpy as np
import tensorflow as tf


def waveform_to_log_mel_spectrogram_patches(waveform, params):
    """Compute log mel spectrogram patches of a 1-D waveform."""
    with tf.name_scope("log_mel_features"):
        # waveform has shape [<# samples>]

        # Convert waveform into spectrogram using a Short-Time Fourier Transform.
        # Note that tf.signal.stft() uses a periodic Hann window by default.
        window_length_samples = int(
            round(params.sample_rate * params.stft_window_seconds)
        )
        hop_length_samples = int(round(params.sample_rate * params.stft_hop_seconds))
        fft_length = 2 ** int(np.ceil(np.log(window_length_samples) / np.log(2.0)))
        num_spectrogram_bins = fft_length // 2 + 1
        if params.tflite_compatible:
            magnitude_spectrogram = _tflite_stft_magnitude(
                signal=waveform,
                frame_length=window_length_samples,
                frame_step=hop_length_samples,
                fft_length=fft_length,
            )
        else:
            magnitude_spectrogram = tf.abs(
                tf.signal.stft(
                    signals=waveform,
                    frame_length=window_length_samples,
                    frame_step=hop_length_samples,
                    fft_length=fft_length,
                )
            )
        # magnitude_spectrogram has shape [<# STFT frames>, num_spectrogram_bins]

        # Convert spectrogram into log mel spectrogram.
        linear_to_mel_weight_matrix = tf.signal.linear_to_mel_weight_matrix(
            num_mel_bins=params.mel_bands,
            num_spectrogram_bins=num_spectrogram_bins,
            sample_rate=params.sample_rate,
            lower_edge_hertz=params.mel_min_hz,
            upper_edge_hertz=params.mel_max_hz,
        )
        mel_spectrogram = tf.matmul(magnitude_spectrogram, linear_to_mel_weight_matrix)
        log_mel_spectrogram = tf.math.log(mel_spectrogram + params.log_offset)
        # log_mel_spectrogram has shape [<# STFT frames>, params.mel_bands]

        # Frame spectrogram (shape [<# STFT frames>, params.mel_bands]) into patches
        # (the input examples). Only complete frames are emitted, so if there is
        # less than params.patch_window_seconds of waveform then nothing is emitted
        # (to avoid this, zero-pad before processing).
        spectrogram_hop_length_samples = int(
            round(params.sample_rate * params.stft_hop_seconds)
        )
        spectrogram_sample_rate = params.sample_rate / spectrogram_hop_length_samples
        patch_window_length_samples = int(
            round(spectrogram_sample_rate * params.patch_window_seconds)
        )
        patch_hop_length_samples = int(
            round(spectrogram_sample_rate * params.patch_hop_seconds)
        )
        features = tf.signal.frame(
            signal=log_mel_spectrogram,
            frame_length=patch_window_length_samples,
            frame_step=patch_hop_length_samples,
            axis=0,
        )
        # features has shape [<# patches>, <# STFT frames in an patch>, params.mel_bands]

        return log_mel_spectrogram, features


def pad_waveform(waveform, params):
    """Pads waveform with silence if needed to get an integral number of patches."""
    # In order to produce one patch of log mel spectrogram input to YAMNet, we
    # need at least one patch window length of waveform plus enough extra samples
    # to complete the final STFT analysis window.
    min_waveform_seconds = (
        params.patch_window_seconds
        + params.stft_window_seconds
        - params.stft_hop_seconds
    )
    min_num_samples = tf.cast(min_waveform_seconds * params.sample_rate, tf.int32)
    num_samples = tf.shape(waveform)[0]
    num_padding_samples = tf.maximum(0, min_num_samples - num_samples)

    # In addition, there might be enough waveform for one or more additional
    # patches formed by hopping forward. If there are more samples than one patch,
    # round up to an integral number of hops.
    num_samples = tf.maximum(num_samples, min_num_samples)
    num_samples_after_first_patch = num_samples - min_num_samples
    hop_samples = tf.cast(params.patch_hop_seconds * params.sample_rate, tf.int32)
    num_hops_after_first_patch = tf.cast(
        tf.math.ceil(
            tf.cast(num_samples_after_first_patch, tf.float32)
            / tf.cast(hop_samples, tf.float32)
        ),
        tf.int32,
    )
    num_padding_samples += (
        hop_samples * num_hops_after_first_patch - num_samples_after_first_patch
    )

    padded_waveform = tf.pad(
        waveform, [[0, num_padding_samples]], mode="CONSTANT", constant_values=0.0
    )
    return padded_waveform


def _tflite_stft_magnitude(signal, frame_length, frame_step, fft_length):
    """TF-Lite-compatible version of tf.abs(tf.signal.stft())."""

    def _hann_window():
        return tf.reshape(
            tf.constant(
                (
                    0.5
                    - 0.5 * np.cos(2 * np.pi * np.arange(0, 1.0, 1.0 / frame_length))
                ).astype(np.float32),
                name="hann_window",
            ),
            [1, frame_length],
        )

    def _dft_matrix(dft_length):
        """Calculate the full DFT matrix in NumPy."""
        # See https://en.wikipedia.org/wiki/DFT_matrix
        omega = (0 + 1j) * 2.0 * np.pi / float(dft_length)
        # Don't include 1/sqrt(N) scaling, tf.signal.rfft doesn't apply it.
        return np.exp(omega * np.outer(np.arange(dft_length), np.arange(dft_length)))

    def _rdft(framed_signal, fft_length):
        """Implement real-input Discrete Fourier Transform by matmul."""
        # We are right-multiplying by the DFT matrix, and we are keeping only the
        # first half ("positive frequencies").  So discard the second half of rows,
        # but transpose the array for right-multiplication.  The DFT matrix is
        # symmetric, so we could have done it more directly, but this reflects our
        # intention better.
        complex_dft_matrix_kept_values = _dft_matrix(fft_length)[
            : (fft_length // 2 + 1), :
        ].transpose()
        real_dft_matrix = tf.constant(
            np.real(complex_dft_matrix_kept_values).astype(np.float32),
            name="real_dft_matrix",
        )
        imag_dft_matrix = tf.constant(
            np.imag(complex_dft_matrix_kept_values).astype(np.float32),
            name="imaginary_dft_matrix",
        )
        signal_frame_length = tf.shape(framed_signal)[-1]
        half_pad = (fft_length - signal_frame_length) // 2
        padded_frames = tf.pad(
            framed_signal,
            [
                # Don't add any padding in the frame dimension.
                [0, 0],
                # Pad before and after the signal within each frame.
                [half_pad, fft_length - signal_frame_length - half_pad],
            ],
            mode="CONSTANT",
            constant_values=0.0,
        )
        real_stft = tf.matmul(padded_frames, real_dft_matrix)
        imag_stft = tf.matmul(padded_frames, imag_dft_matrix)
        return real_stft, imag_stft

    def _complex_abs(real, imag):
        return tf.sqrt(tf.add(real * real, imag * imag))

    framed_signal = tf.signal.frame(signal, frame_length, frame_step)
    windowed_signal = framed_signal * _hann_window()
    real_stft, imag_stft = _rdft(windowed_signal, fft_length)
    stft_magnitude = _complex_abs(real_stft, imag_stft)
    return stft_magnitude



from typing import Dict, Optional, Tuple, Union

import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader
from torch.utils.data import Dataset as TorchDataset

from ertk.dataset import Dataset
from ertk.pytorch.utils import MTLTaskConfig


class DataModuleAdapter(pl.LightningDataModule):
    train_dataset: TorchDataset
    val_dataset: TorchDataset
    test_dataset: TorchDataset

    def __init__(
        self,
        dataset: Dataset,
        train_select: Union[str, Dict[str, str]],
        val_select: Union[str, Dict[str, str]],
        test_select: Union[str, Dict[str, str], None] = None,
        batch_size: int = 32,
        dl_num_workers: int = 0,
    ):
        super().__init__()

        self.dataset = dataset
        self.batch_size = batch_size
        self.train_select = train_select
        self.val_select = val_select
        self.test_select = test_select or val_select
        self.dl_num_workers = dl_num_workers

    def setup(self, stage: Optional[str] = None) -> None:
        self.train_idx = self.dataset.get_idx_for_split(self.train_select)
        self.val_idx = self.dataset.get_idx_for_split(self.val_select)
        self.test_idx = self.dataset.get_idx_for_split(self.test_select)

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.dl_num_workers,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.dl_num_workers,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.dl_num_workers,
        )

    def predict_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.dl_num_workers,
        )


class MTLDataModule(DataModuleAdapter):
    def __init__(
        self,
        dataset: Dataset,
        tasks: Dict[str, MTLTaskConfig],
        train_select: Union[str, Dict[str, str]],
        val_select: Union[str, Dict[str, str]],
        test_select: Union[str, Dict[str, str], None] = None,
        batch_size: int = 32,
        dl_num_workers: int = 0,
    ):
        super().__init__(
            dataset,
            train_select=train_select,
            val_select=val_select,
            test_select=test_select,
            batch_size=batch_size,
            dl_num_workers=dl_num_workers,
        )
        self.tasks = tasks

    def setup(self, stage: Optional[str] = None) -> None:
        super().setup(stage)

        for name in ["train", "val", "test"]:
            idx = getattr(self, f"{name}_idx")
            x = torch.as_tensor(self.dataset.x[idx])
            ys = {
                k: torch.as_tensor(self.dataset.get_group_indices(k)[idx])
                for k in self.tasks
            }
            setattr(self, f"{name}_dataset", MTLDataset(x, ys))


class MTLDataset(TorchDataset[Tuple[torch.Tensor, Dict[str, torch.Tensor]]]):
    def __init__(self, x: torch.Tensor, y: Dict[str, torch.Tensor]):
        self.x = x
        self.y = y

    def __getitem__(self, index):
        return self.x[index], {k: v[index] for k, v in self.y.items()}

    def __len__(self):
        return self.x.size(0)


def create_mtl_dataloader(
    data: Dataset,
    tasks: Dict[str, MTLTaskConfig],
    batch_size: int = 32,
    shuffle: bool = False,
) -> DataLoader[Tuple[torch.Tensor, Dict[str, torch.Tensor]]]:
    x = torch.as_tensor(data.x)
    task_data = {k: torch.as_tensor(data.get_group_indices(k)) for k in tasks}
    dataset = MTLDataset(x, task_data)
    return DataLoader(dataset, batch_size, shuffle=shuffle)



# Copyright 2017 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Post-process embeddings from VGGish."""

import numpy as np

from . import vggish_params


class Postprocessor(object):
    """Post-processes VGGish embeddings.

    The initial release of AudioSet included 128-D VGGish embeddings for each
    segment of AudioSet. These released embeddings were produced by applying
    a PCA transformation (technically, a whitening transform is included as well)
    and 8-bit quantization to the raw embedding output from VGGish, in order to
    stay compatible with the YouTube-8M project which provides visual embeddings
    in the same format for a large set of YouTube videos. This class implements
    the same PCA (with whitening) and quantization transformations.
    """

    def __init__(self, pca_params_npz_path):
        """Constructs a postprocessor.

        Args:
          pca_params_npz_path: Path to a NumPy-format .npz file that
            contains the PCA parameters used in postprocessing.
        """
        params = np.load(pca_params_npz_path)
        self._pca_matrix = params[vggish_params.PCA_EIGEN_VECTORS_NAME]
        # Load means into a column vector for easier broadcasting later.
        self._pca_means = params[vggish_params.PCA_MEANS_NAME].reshape(-1, 1)
        assert self._pca_matrix.shape == (
            vggish_params.EMBEDDING_SIZE,
            vggish_params.EMBEDDING_SIZE,
        ), "Bad PCA matrix shape: %r" % (self._pca_matrix.shape,)
        assert self._pca_means.shape == (
            vggish_params.EMBEDDING_SIZE,
            1,
        ), "Bad PCA means shape: %r" % (self._pca_means.shape,)

    def postprocess(self, embeddings_batch):
        """Applies postprocessing to a batch of embeddings.

        Args:
          embeddings_batch: An nparray of shape [batch_size, embedding_size]
            containing output from the embedding layer of VGGish.

        Returns:
          An nparray of the same shape as the input but of type uint8,
          containing the PCA-transformed and quantized version of the input.
        """
        assert len(embeddings_batch.shape) == 2, "Expected 2-d batch, got %r" % (
            embeddings_batch.shape,
        )
        assert (
            embeddings_batch.shape[1] == vggish_params.EMBEDDING_SIZE
        ), "Bad batch shape: %r" % (embeddings_batch.shape,)

        # Apply PCA.
        # - Embeddings come in as [batch_size, embedding_size].
        # - Transpose to [embedding_size, batch_size].
        # - Subtract pca_means column vector from each column.
        # - Premultiply by PCA matrix of shape [output_dims, input_dims]
        #   where both are are equal to embedding_size in our case.
        # - Transpose result back to [batch_size, embedding_size].
        pca_applied = np.dot(self._pca_matrix, (embeddings_batch.T - self._pca_means)).T

        # Quantize by:
        # - clipping to [min, max] range
        clipped_embeddings = np.clip(
            pca_applied, vggish_params.QUANTIZE_MIN_VAL, vggish_params.QUANTIZE_MAX_VAL
        )
        # - convert to 8-bit in range [0.0, 255.0]
        quantized_embeddings = (clipped_embeddings - vggish_params.QUANTIZE_MIN_VAL) * (
            255.0 / (vggish_params.QUANTIZE_MAX_VAL - vggish_params.QUANTIZE_MIN_VAL)
        )
        # - cast 8-bit float to uint8
        quantized_embeddings = quantized_embeddings.astype(np.uint8)

        return quantized_embeddings



import pickle
import warnings
from pathlib import Path
from typing import Tuple

import click
from sklearn.model_selection import GridSearchCV, GroupKFold
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from ertk.classification import get_balanced_sample_weights
from ertk.config import get_arg_mapping
from ertk.dataset import load_multiple
from ertk.sklearn.models import get_sk_model
from ertk.transform import SequenceTransformWrapper


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option("--features", required=True, help="Features to load.")
@click.option("--clf", "clf_type", required=True, help="Classifier to use.")
@click.option("--save", type=Path, required=True, help="Location to save the model.")
@click.option(
    "--cv",
    type=click.Choice(["speaker", "corpus"]),
    help="Cross-validation method.",
)
@click.option("--balanced/--imbalanced", default=True, help="Balances sample weights.")
@click.option(
    "--normalise",
    type=click.Choice(["speaker", "corpus", "online"]),
    default="speaker",
    help="Normalisation method.",
)
@click.option(
    "--transform",
    type=click.Choice(["std", "minmax"]),
    default="std",
    show_default=True,
    help="Transformation class.",
)
@click.option("--subset", multiple=True, default=["default"], help="Subset selection.")
@click.option("--map_groups", multiple=True, help="Group name mapping.")
@click.option(
    "--sel_groups",
    multiple=True,
    help="Group selection. This is a map from partition to group(s).",
)
@click.option(
    "--clip_seq", type=int, help="Clip sequences to this length (before pad)."
)
@click.option(
    "--pad_seq", type=int, help="Pad sequences to multiple of this length (after clip)."
)
@click.option("--target", default="label", help="Classifier/regressor target.")
@click.option(
    "--clf_args",
    "clf_args_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File containing keyword arguments to give to model initialisation.",
)
@click.option(
    "--param_grid",
    "param_grid_file",
    type=click.Path(exists=True, path_type=Path),
    help="File with parameter grid data.",
)
@click.option("--verbose", count=True, help="Verbose training.")
def main(
    input: Tuple[Path],
    features: str,
    clf_type: str,
    save: Path,
    cv: str,
    normalise: str,
    transform: str,
    balanced: bool,
    subset: Tuple[str],
    map_groups: Tuple[str],
    sel_groups: Tuple[str],
    clip_seq: int,
    pad_seq: int,
    target: str,
    clf_args_file: Path,
    param_grid_file: Path,
    verbose: int,
):
    """Trains a model on the given INPUT datasets. INPUT files must be
    corpus info files.

    Optionally pickles the model.
    """

    dataset = load_multiple(input, features)

    if len(subset) == 1:
        dataset = load_multiple(input, features, subsets=subset[0])
    else:
        subset_map = {}
        for m in subset:
            subset_map.update(get_arg_mapping(m))
        dataset = load_multiple(input, features, subsets=subset_map)

    for m in map_groups:
        mapping = get_arg_mapping(m)
        for part in mapping:
            if part not in dataset.partitions:
                warnings.warn(
                    f"Partition {part} cannot be mapped as it is not in the dataset."
                )
                continue
            dataset.map_groups(part, mapping[part])
    for m in sel_groups:
        mapping = get_arg_mapping(m)
        for part in mapping:
            if part not in dataset.partitions:
                warnings.warn(
                    f"Partition {part} cannot be selected as it is not in the dataset."
                )
                continue
            keep = mapping[part]
            if isinstance(keep, str):
                keep = [keep]
            dataset.remove_groups(part, keep=keep)

    if clip_seq:
        dataset.clip_arrays(clip_seq)
    if pad_seq:
        dataset.pad_arrays(pad_seq)

    transformer = {"std": StandardScaler, "minmax": MinMaxScaler}[transform]()
    if normalise == "none":
        transformer = None
    elif normalise == "online" and len(dataset.x[0].shape) > 1:
        transformer = SequenceTransformWrapper(transformer, "feature")
    elif normalise != "online":
        dataset.normalise(normaliser=transformer, partition=normalise)
        transformer = None

    print(dataset)

    sample_weight = None
    if balanced and target in dataset.partitions:
        sample_weight = get_balanced_sample_weights(dataset.get_annotations(target))
    fit_params = {"sample_weight": sample_weight}

    model_args = {}
    if clf_args_file:
        model_args = get_arg_mapping(clf_args_file)
    param_grid = {}
    if param_grid_file:
        param_grid = get_arg_mapping(param_grid_file)

    clf_lib, clf_type = clf_type.split("/")
    if clf_lib == "sk":
        model_args["probability"] = True
        clf = get_sk_model(clf_type, **model_args)
        clf = Pipeline([("transform"), ("clf", clf)])
        if cv:
            groups = dataset.get_group_indices(cv)
            clf = GridSearchCV(
                clf,
                param_grid,
                scoring="balanced_accuracy",
                cv=GroupKFold(2),
                n_jobs=-1,
                verbose=verbose,
            )
            fit_params["groups"] = groups

        y = dataset.get_annotations(target)
        clf.fit(dataset.x, y, **fit_params)
        save.parent.mkdir(parents=True, exist_ok=True)
        with open(save, "wb") as fid:
            pickle.dump(clf.best_estimator_, fid)
            print(f"Saved classifier to {save}")


if __name__ == "__main__":
    main()



from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import tensorflow as tf
from sklearn.pipeline import Pipeline
from tensorflow.keras.layers import Layer, Wrapper
from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam, Optimizer
from tensorflow.keras.utils import Sequence

from ertk.utils import batch_arrays_by_length, shuffle_multiple

TFModelFunction = Callable[..., Union[Model, Pipeline]]
DataFunction = Callable[..., tf.data.Dataset]


def compile_wrap(
    model_fn: Optional[TFModelFunction] = None,
    opt_cls: Type[Optimizer] = Adam,
    opt_kwargs: Dict[str, Any] = dict(learning_rate=0.0001),
    metrics: List[Union[str, Metric]] = ["sparse_categorical_accuracy"],
    loss: Union[str, Loss] = "sparse_categorical_crossentropy",
    **compile_kwargs,
):
    """Wrapper that takes a model creation function and gives a new
    function which returns a compiled model with the given compile
    parameters.

    Parameters
    ----------
    model_fn: callable, optional
        A method that returns an uncompiled model.
    opt_cls: type
        The Optimizer class to use.
    opt_kwargs: dict
        Keyword arguments to pass to opt_cls.
    metrics: list
        List of metrics to use.
    loss: Loss
        The loss function to use.
    **compile_kwargs: dict
        Other keyword arguments to pass to the model's compile() method.
    """

    def _wrapper(func: Callable[..., Model]):
        @wraps(func)
        def new_model_fn(*args, **kwargs) -> Model:
            model = func(*args, **kwargs)
            model.compile(
                optimizer=opt_cls(**opt_kwargs),
                metrics=metrics,
                loss=loss,
                **compile_kwargs,
            )
            return model

        return new_model_fn

    if model_fn is not None:
        return _wrapper(model_fn)

    return _wrapper


def test_fit(
    model_fn: TFModelFunction,
    input_size: Tuple[int, ...],
    *args,
    batch_size: int = 64,
    num_instances: int = 7000,
    **kwargs,
):
    """Tests the given model architecture/structure by training it on
    dummy data.

    Parameters
    ----------
    model_fn: callable
        Function that returns a Keras model. Called as model_fn(*args,
        **kwargs).
    input_size: tuple of int
        Input shape to the model. This is used to generate dummy data of
        the correct shape.
    *args
        Positional arguments to pass to model_fn().
    batch_size: int
        The batch size to use.
    num_instances: int
        The number of instances to generate.
    **kwargs
        Keyword arguments to pass to model_fn().
    """
    for gpu in tf.config.get_visible_devices("GPU"):
        tf.config.experimental.set_memory_growth(gpu, True)

    compiled_fn = compile_wrap(model_fn)
    model = compiled_fn(*args, n_classes=7, **kwargs)
    model.summary()

    valid = num_instances // 10
    rng = np.random.default_rng()
    x = rng.normal(size=(num_instances,) + input_size)
    y = rng.integers(7, size=num_instances)
    train_data = tf.data.Dataset.from_tensor_slices((x[valid:], y[valid:]))
    train_data = train_data.batch(batch_size)
    valid_data = tf.data.Dataset.from_tensor_slices((x[:valid], y[:valid]))
    valid_data = valid_data.batch(batch_size)
    model.fit(train_data, validation_data=valid_data, epochs=2, verbose=1)


def tf_dataset_gen(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    *,
    batch_size: int = 64,
    shuffle: bool = True,
):
    """Returns a TensorFlow generator Dataset instance with the given
    data.

    Parameters
    ----------
    x: numpy.ndarray
        A 2- or 3-D data matrix of shape (n_instances, n_features) or
        (n_instances, seq_len, n_features).
    y: numpy.ndarray
        A 1-D array of length n_instances containing numeric class
        labels.
    sample_weight: numpy.ndarray, optional
        A 1-D array of length n_instances containing sample weights.
        Added as third item in dataset if present.
    batch_size: int
        The batch size to use.
    shuffle: boolean
        Whether or not to shuffle the dataset. Note that shuffling is
        done *before* batching, unlike in `create_tf_dataset_ragged()`.
    """

    def gen_inst():
        if shuffle:
            perm = np.random.permutation(len(x))
        else:
            perm = np.arange(len(x))

        if sample_weight is None:
            for i in perm:
                yield x[i], y[i]
        else:
            for i in perm:
                yield x[i], y[i], sample_weight[i]

    sig: Tuple[tf.TensorSpec, ...] = (
        tf.TensorSpec(shape=x[0].shape, dtype=tf.float32),
        tf.TensorSpec(shape=(), dtype=tf.int64),
    )
    if sample_weight is not None:
        sig += (tf.TensorSpec(shape=(), dtype=tf.float32),)
    data = tf.data.Dataset.from_generator(gen_inst, output_signature=sig)
    return data.batch(batch_size).prefetch(2)


def tf_dataset_mem(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    *,
    batch_size: int = 64,
    shuffle: bool = True,
) -> tf.data.Dataset:
    """Returns a TensorFlow in-memory Dataset instance with the given
    data.

    Parameters
    ----------
    x: numpy.ndarray
        A 2- or 3-D data matrix of shape (n_instances, n_features) or
        (n_instances, seq_len, n_features).
    y: numpy.ndarray
        A 1-D array of length n_instances containing numeric class
        labels.
    sample_weight: numpy.ndarray, optional
        A 1-D array of length n_instances containing sample weights.
        Added as third item in dataset if present.
    batch_size: int
        The batch size to use.
    shuffle: boolean
        Whether or not to shuffle the dataset. Note that shuffling is
        done *before* batching, unlike in `create_tf_dataset_ragged()`.
    """
    with tf.device("CPU"):
        if sample_weight is None:
            data = tf.data.Dataset.from_tensor_slices((x, y))
        else:
            data = tf.data.Dataset.from_tensor_slices((x, y, sample_weight))

    if shuffle:
        data = data.shuffle(len(x))
    return data.batch(batch_size).prefetch(2)


def tf_dataset_mem_ragged(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    *,
    batch_size: int = 64,
    shuffle: bool = True,
) -> tf.data.Dataset:
    """Returns a TensorFlow in-memory Dataset instance from
    variable-length features.

    Parameters
    ----------
    x: numpy.ndarray
        A 3-D data matrix of shape (n_instances, length[i], n_features)
        with variable length axis 1.
    y: numpy.ndarray
        A 1-D array of length n_instances containing numeric class
        labels.
    sample_weight: numpy.ndarray, optional
        A 1-D array of length n_instances containing sample weights.
        Added as third item in dataset if present.
    batch_size: int
        The batch size to use.
    shuffle: boolean
        Whether or not to shuffle the dataset. Note that shuffling is
        done *after* batching, because sequences are sorted by length,
        then batched in similar lengths.
    """

    def ragged_to_dense(x: tf.RaggedTensor, y):
        return x.to_tensor(), y

    def ragged_to_dense_weighted(x: tf.RaggedTensor, y, sample_weight):
        return x.to_tensor(), y, sample_weight

    # Sort according to length
    perm = np.argsort([len(a) for a in x])
    x = x[perm]
    y = y[perm]
    if sample_weight is not None:
        sample_weight = sample_weight[perm]

    ragged = tf.RaggedTensor.from_row_lengths(
        np.concatenate(list(x)), [len(a) for a in x]
    )
    with tf.device("CPU"):
        if sample_weight is None:
            data = tf.data.Dataset.from_tensor_slices((ragged, y))
        else:
            data = tf.data.Dataset.from_tensor_slices((ragged, y, sample_weight))

    # Group similar lengths in batches, then shuffle batches
    data = data.batch(batch_size)
    if shuffle:
        data = data.shuffle(len(x) // batch_size + 1)

    if sample_weight is None:
        data = data.map(ragged_to_dense)
    else:
        data = data.map(ragged_to_dense_weighted)
    return data.prefetch(2)


class BatchedFrameSequence(Sequence):
    """Creates a sequence of batches of frames to process.

    Parameters
    ----------
    x: ndarray or list of ndarray
        Sequences of vectors.
    y: ndarray
        Labels corresponding to sequences in x.
    prebatched: bool, default = False
        Whether or not x has already been grouped into batches.
    batch_size: int, default = 32
        Batch size to use. Each generated batch will be at most this size.
    shuffle: bool, default = True
        Whether to shuffle the order of the batches.
    """

    def __init__(
        self,
        x: Union[np.ndarray, List[np.ndarray]],
        y: np.ndarray,
        prebatched: bool = False,
        batch_size: int = 32,
        shuffle: bool = True,
    ):
        self.x = x
        self.y = y
        if not prebatched:
            self.x, self.y = batch_arrays_by_length(
                self.x, self.y, batch_size=batch_size, shuffle=shuffle
            )
        if shuffle:
            self.x, self.y = shuffle_multiple(self.x, self.y, numpy_indexing=True)

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx: int):
        return self.x[idx], self.y[idx]


class BatchedSequence(Sequence):
    """Creates a sequence of batches to process.

    Parameters
    ----------
    x: ndarray or list of ndarray
        Instance feature vectors. Each vector is assumed to be for a different
        instance.
    y: ndarray
        Labels corresponding to sequences in x.
    prebatched: bool, default = False
        Whether or not x has already been grouped into batches.
    batch_size: int, default = 32
        Batch size to use. Each generated batch will be at most this size.
    shuffle: bool, default = True
        Whether to shuffle the instances.
    """

    def __init__(
        self, x: np.ndarray, y: np.ndarray, batch_size: int = 32, shuffle: bool = True
    ):
        self.x = x
        self.y = y
        self.batch_size = batch_size
        if shuffle:
            self.x, self.y = shuffle_multiple(self.x, self.y, numpy_indexing=True)

    def __len__(self):
        return int(np.ceil(len(self.x) / self.batch_size))

    def __getitem__(self, idx: int):
        sl = slice(idx * self.batch_size, (idx + 1) * self.batch_size)
        return self.x[sl], self.y[sl]


def print_linear_model_structure(model: Model):
    """Prints the structure of a "sequential" model by listing the layer
    types and shapes in order.

    Parameters
    ----------
    model: Model
        The model to describe.
    """

    def print_inner(model: Layer, depth: int = 0):
        indent = "\t" * depth
        if not isinstance(model, Model):
            print(indent, model.name, model.output_shape)
            return

        for layer in model.layers:
            name = layer.name
            if name.startswith("tf_op_layer_"):
                name = name[12:]

            print(indent, name, layer.output_shape)
            if isinstance(layer, Model):
                print_inner(layer, depth + 1)
            elif isinstance(layer, Wrapper):
                print_inner(layer.layer, depth + 1)

    print_inner(model)


def init_gpu_memory_growth():
    """Sets TensorFlow to allocate memory on GPU as needed instead of
    all at once.
    """
    for gpu in tf.config.list_physical_devices("GPU"):
        tf.config.experimental.set_memory_growth(gpu, True)



from dataclasses import dataclass
from types import ModuleType
from typing import Iterable, List, Optional, Union

import numpy as np
from omegaconf import MISSING
from scipy.ndimage import zoom
from tensorflow.keras.applications import (
    densenet,
    efficientnet,
    inception_resnet_v2,
    inception_v3,
    mobilenet,
    mobilenet_v2,
    mobilenet_v3,
    nasnet,
    resnet,
    vgg16,
    vgg19,
    xception,
)

from ertk.config import ERTKConfig
from ertk.preprocessing.spectrogram import spectrogram
from ertk.tensorflow.utils import init_gpu_memory_growth
from ertk.utils import is_mono_audio

from .base import AudioClipProcessor, FeatureExtractor


@dataclass
class ModelArgs:
    module: ModuleType
    funcname: str
    size: int


MODEL_MAP = {
    "densenet121": ModelArgs(densenet, "DenseNet121", 224),
    "densenet169": ModelArgs(densenet, "DenseNet169", 224),
    "densenet201": ModelArgs(densenet, "DenseNet201", 224),
    "efficientnet_b0": ModelArgs(efficientnet, "EfficientNetB0", 224),
    "efficientnet_b1": ModelArgs(efficientnet, "EfficientNetB1", 240),
    "efficientnet_b2": ModelArgs(efficientnet, "EfficientNetB2", 260),
    "efficientnet_b3": ModelArgs(efficientnet, "EfficientNetB3", 300),
    "efficientnet_b4": ModelArgs(efficientnet, "EfficientNetB4", 380),
    "efficientnet_b5": ModelArgs(efficientnet, "EfficientNetB5", 456),
    "efficientnet_b6": ModelArgs(efficientnet, "EfficientNetB6", 528),
    "efficientnet_b7": ModelArgs(efficientnet, "EfficientNetB7", 600),
    "inception_resnet_v2": ModelArgs(inception_resnet_v2, "InceptionResNetV2", 299),
    "inception_v3": ModelArgs(inception_v3, "InceptionV3", 299),
    "mobilenet": ModelArgs(mobilenet, "MobileNet", 224),
    "mobilenet_v2": ModelArgs(mobilenet_v2, "MobileNetV2", 224),
    "mobilenet_v3_large": ModelArgs(mobilenet_v3, "MobileNetV3Large", 224),
    "mobilenet_v3_small": ModelArgs(mobilenet_v3, "MobileNetV3Small", 224),
    "nasnet_large": ModelArgs(nasnet, "NASNetLarge", 331),
    "nasnet_mobile": ModelArgs(nasnet, "NASNetMobile", 224),
    "resnet50": ModelArgs(resnet, "ResNet50", 224),
    "resnet50_v2": ModelArgs(resnet, "ResNet50V2", 224),
    "resnet101": ModelArgs(resnet, "ResNet101", 224),
    "resnet101_v2": ModelArgs(resnet, "ResNet101V2", 224),
    "resnet152": ModelArgs(resnet, "ResNet152", 224),
    "resnet152_v2": ModelArgs(resnet, "ResNet152V2", 224),
    "vgg16": ModelArgs(vgg16, "VGG16", 224),
    "vgg19": ModelArgs(vgg19, "VGG19", 224),
    "xception": ModelArgs(xception, "Xception", 299),
}


@dataclass
class KerasAppsExtractorConfig(ERTKConfig):
    model: str = MISSING
    size: Optional[int] = None
    n_mels: int = 128
    spec_type: str = "mel"
    fmax: int = 8000


class KerasAppsExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="keras_apps",
    config=KerasAppsExtractorConfig,
):
    config: KerasAppsExtractorConfig

    def __init__(self, config: KerasAppsExtractorConfig) -> None:
        super().__init__(config)

        init_gpu_memory_growth()

        args = {"include_top": False, "weights": "imagenet", "pooling": "avg"}
        model_args = MODEL_MAP[config.model]
        size = config.size or model_args.size
        self.size = size
        args["input_shape"] = (size, size, 3)
        func = getattr(model_args.module, model_args.funcname)
        self.model = func(**args)
        self.preprocess_input = getattr(model_args.module, "preprocess_input")

    def _spectrogram(self, x: np.ndarray, sr: float) -> np.ndarray:
        return spectrogram(
            x,
            sr=16000,
            kind=self.config.spec_type,
            pre_emphasis=0,
            window_size=0.025,
            window_shift=0.01,
            n_mels=self.config.n_mels,
            fmin=0,
            fmax=None,
            power=2,
            to_db=True,
        )

    def _to_img(self, spec: np.ndarray) -> np.ndarray:
        scale = self.size / np.array(spec.shape)
        spec = zoom(spec, scale)
        spec = np.interp(spec, (spec.min(), spec.max()), (0, 255))
        return np.tile(spec[:, :, None], (1, 1, 3))

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if is_mono_audio(x):
            x = self._spectrogram(x, kwargs.pop("sr"))
        x = self._to_img(x)
        x = self.preprocess_input(x)
        return self.model(x).numpy()

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        batch = list(batch)
        if all(is_mono_audio(x) for x in batch):
            sr = kwargs.pop("sr")
            batch = [self._spectrogram(x, sr) for x in batch]
        batch = [self._to_img(x) for x in batch]
        batch = np.stack(batch)
        batch = self.preprocess_input(batch)
        return list(self.model(batch).numpy())

    @property
    def dim(self) -> int:
        return super().dim

    @property
    def is_sequence(self) -> bool:
        return False



from pathlib import Path
from typing import List, Tuple

import click

from ertk.dataset import read_features, write_features


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.argument("output", type=Path)
@click.option(
    "--prefix_corpus", is_flag=True, help="Prefix corpus names to instance names."
)
def main(input: Tuple[Path], output: Path, prefix_corpus: bool):
    """Combines multiple INPUT features and writes to OUTPUT."""

    if len(input) == 0:
        raise ValueError("No input files specified.")

    feature_names: List[str] = []
    features = []
    names = []
    for filename in input:
        data = read_features(filename)
        if len(feature_names) == 0:
            feature_names = list(data.feature_names)
        elif len(data.feature_names) != len(feature_names):
            raise ValueError("Feature size of all datasets must match.")
        features += list(data.features)
        if prefix_corpus:
            names += [f"{data.corpus}_{x}" for x in data.names]
        else:
            names += data.names
    if len(set(names)) != len(features):
        raise ValueError(
            "Some names are not unique, use --prefix_corpus to generate unique names."
        )

    output.parent.mkdir(parents=True, exist_ok=True)
    write_features(
        output,
        features,
        corpus="combined",
        names=names,
        feature_names=feature_names,
    )
    print(f"Wrote combined features to {output}")


if __name__ == "__main__":
    main()



from .annotation import read_annotations, write_annotations
from .dataset import CombinedDataset, Dataset, LabelledDataset, load_multiple
from .features import read_features, register_format, write_features
from .utils import (
    get_audio_paths,
    resample_audio,
    resample_rename_clips,
    write_filelist,
)



from typing import Sequence

from torch import nn


class Model(nn.Module):
    def __init__(
        self,
        n_features: int,
        n_classes: int,
        units: Sequence[int] = [512],
        dropout: float = 0.5,
    ):
        super().__init__()

        sizes = [n_features] + list(units)
        self.hidden = nn.Sequential(
            *(
                nn.Sequential(
                    nn.Linear(sizes[i], sizes[i + 1]), nn.ReLU(), nn.Dropout(dropout)
                )
                for i in range(len(units))
            )
        )
        self.final = nn.Linear(units[-1], n_classes)

    def forward(self, x):
        return self.final(self.hidden(x))



from pathlib import Path

import click

from ertk.dataset import read_features


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("output", type=Path)
@click.option("--corpus", type=str, help="Corpus attribute to set, if required.")
@click.option(
    "--header/--noheader",
    default=True,
    help="Input CSV has header/Write output CSV header.",
    show_default=True,
)
@click.option(
    "--label/--nolabel",
    default=False,
    help="Input has label column.",
    show_default=True,
)
def main(input: Path, output: Path, corpus: str, header: bool, label: bool):
    """Convert INPUT dataset format to OUTPUT format. Note that no label
    information is written to OUTPUT.
    """

    if input.suffix == output.suffix:
        raise ValueError("Input format must be different to output.")

    print(f"Reading {input}")
    data = read_features(input, header=header, label=label)
    if corpus:
        data._corpus = corpus
    output.parent.mkdir(parents=True, exist_ok=True)
    data.write(output, header=header)
    print(f"Wrote dataset to {output}")


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import Callable, Optional, Tuple

import pytorch_lightning as pl
import torch
import torch.nn as nn


class SimpleModel(pl.LightningModule, ABC):
    """A model that takes input x, gives output y, and has a single loss
    function. Each batch is either (x, y) pairs or (x, y, sw) triplets,
    where sw is the sample weights.

    Parameters
    ----------
    n_features: int
        Input dimensionality.
    loss: callable
        A callable that takes predictions yhat and ground truth y to
        yield a loss::
            l = loss(yhat, y)
        It is assumed that if sw is given then the returned loss is a 1D
        tensor of losses per instance in the batch.
    optim_fn: callable
        Function that returns an optimiser for use with this model. The
        default optimiser is `torch.optim.Adam` with default args.
    """

    def __init__(
        self,
        n_features: int,
        loss: Callable[..., torch.Tensor],
        optim_fn: Callable[..., torch.optim.Optimizer] = torch.optim.Adam,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.loss = loss
        self.n_features = n_features
        self.optim_fn = optim_fn

    @abstractmethod
    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        raise NotImplementedError()

    def training_step(self, batch, batch_idx: int) -> torch.Tensor:
        x, y, yhat, sw = self.get_outputs_for_batch(batch)
        loss = self.log_loss("train", y, yhat, sw)
        return loss

    def validation_step(self, batch, batch_idx: int) -> None:
        x, y, yhat, sw = self.get_outputs_for_batch(batch)
        self.log_loss("val", y, yhat, sw)

    def log_loss(
        self, name: str, y: torch.Tensor, yhat: torch.Tensor, sw: Optional[torch.Tensor]
    ) -> torch.Tensor:
        loss = self.loss(yhat, y)
        if sw is not None:
            loss = loss.dot(sw) / sw.sum()
        else:
            loss = loss.mean()
        self.log(f"{name}_loss", loss.item())
        return loss

    def get_outputs_for_batch(
        self, batch
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        x, y, *sw = batch
        sw = sw[0] if len(sw) > 0 else None
        x = self.preprocess_input(x)
        yhat = self.forward(x)
        return x, y, yhat, sw

    def predict_step(self, batch, batch_idx: int):
        x, *_ = batch
        return self.forward(x)

    @abstractmethod
    def preprocess_input(self, x: torch.Tensor) -> torch.Tensor:
        pass

    def configure_optimizers(self):
        return self.optim_fn(self.parameters())


class SimpleClassificationModel(SimpleModel):
    """A classification model that takes input x, gives output
    probabilities y, and has a single loss function. Each batch is
    either (x, y) pairs or (x, y, sw) triplets, where sw is the sample
    weights.

    Parameters
    ----------
    n_classes: int
        Number of output classes.
    kwargs: dict
        Other keyword arguments to pass to `SimpleModel.__init__()`.
    """

    def __init__(self, n_classes: int, **kwargs) -> None:
        super().__init__(loss=nn.CrossEntropyLoss(reduction="none"), **kwargs)
        self.n_classes = n_classes

    def training_step(self, batch, batch_idx: int) -> torch.Tensor:
        x, y, yhat, sw = self.get_outputs_for_batch(batch)
        loss = self.log_loss("train", y, yhat, sw)
        self.log_acc("train", y, yhat)
        return loss

    def validation_step(self, batch, batch_idx: int) -> None:
        x, y, yhat, sw = self.get_outputs_for_batch(batch)
        self.log_loss("val", y, yhat, sw)
        self.log_acc("val", y, yhat)

    def log_acc(self, name: str, y: torch.Tensor, yhat: torch.Tensor) -> torch.Tensor:
        acc = torch.sum(yhat.argmax(1) == y) / float(len(y))
        self.log(f"{name}_acc", acc.item())
        return acc



import xml.etree.ElementTree as ET
from pathlib import Path
from typing import List, Tuple

import click
import librosa
import soundfile
from joblib import delayed

from ertk.utils import TqdmParallel


@click.command()
@click.argument("input", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.argument("output", type=Path)
@click.option("--prefix", type=str, default="")
@click.option(
    "--tier", default="phrase", help="Tier from which to extract transcripts."
)
@click.option(
    "--break_on",
    multiple=True,
    default=["pause", "int"],
    help="Break the utterance at these annotation values.",
)
def main(
    input: Tuple[Path], output: Path, prefix: str, tier: str, break_on: Tuple[str]
):
    """Splits ELAN (.eaf) files and associated audio into segments."""

    def process(path: Path):
        audio, _ = librosa.load(path, sr=16000, res_type="kaiser_fast")

        eaf_file = path.with_suffix(".eaf")
        xml = ET.parse(eaf_file)

        header = xml.find("HEADER")
        if (
            header
            and "TIME_UNITS" in header.attrib
            and header.attrib["TIME_UNITS"] != "milliseconds"
        ):
            raise ValueError("`TIME_UNITS` should be 'milliseconds'.")

        time_order = xml.find("TIME_ORDER")
        if time_order is None:
            raise RuntimeError("EAF document has no `TIME_ORDER` element.")

        time_slots = {}
        for time_slot in time_order:
            time_value = time_slot.attrib.get("TIME_VALUE", "")
            time_slots[time_slot.attrib["TIME_SLOT_ID"]] = time_value

        phrases = next(
            (x for x in xml.iterfind("TIER") if x.attrib["TIER_ID"] == tier),
            None,
        )
        if phrases is None:
            # No linguistic content
            return

        group: List[Tuple[str, str, str]] = []  # [(start, end, word), ...]
        utts: List[Tuple[str, str, str]] = []  # [(start, end, txt), ...]
        for annotation in phrases:
            alignable_annotation = annotation[0]
            ts1 = time_slots[alignable_annotation.attrib["TIME_SLOT_REF1"]]
            ts2 = time_slots[alignable_annotation.attrib["TIME_SLOT_REF2"]]
            annotation_value = alignable_annotation[0].text
            if annotation_value in break_on and len(group) > 0:
                utts.append(
                    (group[0][0], group[-1][1], " ".join([x[2] for x in group]))
                )
                group = []
            elif annotation_value is not None:
                group.append((ts1, ts2, annotation_value))
        if len(group) > 0:
            utts.append((group[0][0], group[-1][1], " ".join([x[2] for x in group])))

        out_txt = output / f"{prefix}{path.stem}.txt"
        with open(out_txt, "w") as fid:
            for i, (start, end, w) in enumerate(utts):
                out_name = f"{prefix}{path.stem}_{i:03d}"
                print(out_name, w)
                fid.write(f"{out_name}\t{w}\n")
                s_sam = int(start) * 16
                e_sam = int(end) * 16
                out_audio = output / f"{out_name}.wav"
                soundfile.write(out_audio, audio[s_sam:e_sam], 16000)

    output.mkdir(parents=True, exist_ok=True)
    all_files = [p for path in input for p in path.glob("**/*.wav")]
    TqdmParallel(total=len(all_files), desc="Processing ELAN files", n_jobs=-1)(
        delayed(process)(p) for p in all_files
    )


if __name__ == "__main__":
    main()



# Copyright 2017 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Defines the 'VGGish' model used to generate AudioSet embedding features.

The public AudioSet release (https://research.google.com/audioset/download.html)
includes 128-D features extracted from the embedding layer of a VGG-like model
that was trained on a large Google-internal YouTube dataset. Here we provide
a TF-Slim definition of the same model, without any dependences on libraries
internal to Google. We call it 'VGGish'.

Note that we only define the model up to the embedding layer, which is the
penultimate layer before the final classifier layer. We also provide various
hyperparameter values (in vggish_params.py) that were used to train this model
internally.

For comparison, here is TF-Slim's VGG definition:
https://github.com/tensorflow/models/blob/master/research/slim/nets/vgg.py
"""

import tensorflow.compat.v1 as tf
import tf_slim as slim

from . import vggish_params as params


def define_vggish_slim(features_tensor=None, training=False):
    """Defines the VGGish TensorFlow model.

    All ops are created in the current default graph, under the scope 'vggish/'.

    The input is either a tensor passed in via the optional 'features_tensor'
    argument or a placeholder created below named 'vggish/input_features'. The
    input is expected to have dtype float32 and shape [batch_size, num_frames,
    num_bands] where batch_size is variable and num_frames and num_bands are
    constants, and [num_frames, num_bands] represents a log-mel-scale spectrogram
    patch covering num_bands frequency bands and num_frames time frames (where
    each frame step is usually 10ms). This is produced by computing the stabilized
    log(mel-spectrogram + params.LOG_OFFSET).  The output is a tensor named
    'vggish/embedding' which produces the pre-activation values of a 128-D
    embedding layer, which is usually the penultimate layer when used as part of a
    full model with a final classifier layer.

    Args:
      features_tensor: If not None, the tensor containing the input features.
        If None, a placeholder input is created.
      training: If true, all parameters are marked trainable.

    Returns:
      The op 'vggish/embeddings'.
    """
    # Defaults:
    # - All weights are initialized to N(0, INIT_STDDEV).
    # - All biases are initialized to 0.
    # - All activations are ReLU.
    # - All convolutions are 3x3 with stride 1 and SAME padding.
    # - All max-pools are 2x2 with stride 2 and SAME padding.
    with slim.arg_scope(
        [slim.conv2d, slim.fully_connected],
        weights_initializer=tf.truncated_normal_initializer(stddev=params.INIT_STDDEV),
        biases_initializer=tf.zeros_initializer(),
        activation_fn=tf.nn.relu,
        trainable=training,
    ), slim.arg_scope(
        [slim.conv2d], kernel_size=[3, 3], stride=1, padding="SAME"
    ), slim.arg_scope(
        [slim.max_pool2d], kernel_size=[2, 2], stride=2, padding="SAME"
    ), tf.variable_scope(
        "vggish"
    ):
        # Input: a batch of 2-D log-mel-spectrogram patches.
        if features_tensor is None:
            features_tensor = tf.placeholder(
                tf.float32,
                shape=(None, params.NUM_FRAMES, params.NUM_BANDS),
                name="input_features",
            )
        # Reshape to 4-D so that we can convolve a batch with conv2d().
        net = tf.reshape(features_tensor, [-1, params.NUM_FRAMES, params.NUM_BANDS, 1])

        # The VGG stack of alternating convolutions and max-pools.
        net = slim.conv2d(net, 64, scope="conv1")
        net = slim.max_pool2d(net, scope="pool1")
        net = slim.conv2d(net, 128, scope="conv2")
        net = slim.max_pool2d(net, scope="pool2")
        net = slim.repeat(net, 2, slim.conv2d, 256, scope="conv3")
        net = slim.max_pool2d(net, scope="pool3")
        net = slim.repeat(net, 2, slim.conv2d, 512, scope="conv4")
        net = slim.max_pool2d(net, scope="pool4")

        # Flatten before entering fully-connected layers
        net = slim.flatten(net)
        net = slim.repeat(net, 2, slim.fully_connected, 4096, scope="fc1")
        # The embedding layer.
        net = slim.fully_connected(
            net, params.EMBEDDING_SIZE, scope="fc2", activation_fn=None
        )
        return tf.identity(net, name="embedding")


def load_vggish_slim_checkpoint(session, checkpoint_path):
    """Loads a pre-trained VGGish-compatible checkpoint.

    This function can be used as an initialization function (referred to as
    init_fn in TensorFlow documentation) which is called in a Session after
    initializating all variables. When used as an init_fn, this will load
    a pre-trained checkpoint that is compatible with the VGGish model
    definition. Only variables defined by VGGish will be loaded.

    Args:
      session: an active TensorFlow session.
      checkpoint_path: path to a file containing a checkpoint that is
        compatible with the VGGish model definition.
    """
    # Get the list of names of all VGGish variables that exist in
    # the checkpoint (i.e., all inference-mode VGGish variables).
    with tf.Graph().as_default():
        define_vggish_slim(training=False)
        vggish_var_names = [v.name for v in tf.global_variables()]

    # Get the list of all currently existing variables that match
    # the list of variable names we just computed.
    vggish_vars = [v for v in tf.global_variables() if v.name in vggish_var_names]

    # Use a Saver to restore just the variables selected above.
    saver = tf.train.Saver(vggish_vars, name="vggish_load_pretrained", write_version=1)
    saver.restore(session, checkpoint_path)



import numpy as np
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import (
    add,
    Dense,
    GlobalAvgPool1D,
    GlobalMaxPool1D,
    Input,
    MultiHeadAttention,
    LayerNormalization,
    ReLU,
)


def _encoder_layer(x, num_heads: int, model_dim: int):
    att = MultiHeadAttention(num_heads=num_heads, key_dim=model_dim)(x, x)
    x = LayerNormalization()(x + att)
    ff = Dense(2048)(x)
    ff = ReLU()(ff)
    ff = Dense(model_dim)(ff)
    return LayerNormalization()(x + ff)


def _positional_encoding(seq_len: int, model_dim: int):
    if model_dim % 2 != 0:
        raise ValueError("model_dim must be divisible by 2.")
    pos = np.arange(seq_len, dtype=np.float32)[:, None]
    dims = np.arange(model_dim // 2, dtype=np.float32)[None, :]
    angles = pos / np.power(10000, 2 * dims / model_dim)
    # Need to include batch dimension
    pe = np.empty((1, seq_len, model_dim))
    pe[..., 0::2] = np.sin(angles)
    pe[..., 1::2] = np.cos(angles)
    return tf.constant(pe, dtype=tf.float32)


def model(
    n_features: int,
    n_classes: int,
    enc_layers: int = 2,
    num_heads: int = 4,
    seq_len: int = 512,
    model_dim: int = 512,
    pool: str = "max",
) -> Model:
    inputs = Input((seq_len, n_features))
    x = Dense(model_dim, name="projection")(inputs)
    x = add([x, _positional_encoding(seq_len, model_dim)], name="positional_encoding")
    for _ in range(enc_layers):
        x = _encoder_layer(x, num_heads=num_heads, model_dim=model_dim)
    if pool == "max":
        x = GlobalMaxPool1D()(x)
    else:
        x = GlobalAvgPool1D()(x)
    x = Dense(n_classes, activation="softmax")(x)
    return Model(inputs=inputs, outputs=x)



import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import (
    RNN,
    AbstractRNNCell,
    Bidirectional,
    Dense,
    GRUCell,
    Input,
    StackedRNNCells,
    concatenate,
)

__all__ = ["audeep_trae"]


def _dropout_gru_cell(units: int = 256, dropout: float = 0.8) -> AbstractRNNCell:
    return tf.nn.RNNCellDropoutWrapper(
        GRUCell(units, name="gru_cell"), 1 - dropout, 1 - dropout
    )


def _make_rnn(
    units: int = 256,
    layers: int = 2,
    bidirectional: bool = False,
    dropout: float = 0.2,
    name="rnn",
) -> RNN:
    cells = [_dropout_gru_cell(units, dropout) for _ in range(layers)]
    stacked_cell = StackedRNNCells(cells)
    rnn = RNN(
        stacked_cell,
        return_sequences=True,
        return_state=True,
        time_major=True,
        name=name,
    )
    if bidirectional:
        rnn = Bidirectional(rnn, name="bidirectional_" + name)
    return rnn


def audeep_trae(
    input_shape: tuple,
    units: int = 256,
    layers: int = 2,
    bidirectional_encoder: bool = False,
    bidirectional_decoder: bool = False,
) -> Model:
    inputs = Input(input_shape)

    trans = tf.transpose(inputs, [1, 0, 2])

    # Make encoder layers
    encoder = _make_rnn(units, layers, bidirectional_encoder, name="encoder")
    enc_states = encoder(trans)[1:]
    encoder_output = concatenate(enc_states, name="encoder_output_state")

    # Fully connected needs to have output dimension equal to dimension of
    # decoder state
    dec_dim = units * layers
    if bidirectional_decoder:
        dec_dim *= 2
    representation = Dense(dec_dim, activation="tanh", name="representation")(
        encoder_output
    )

    # Initial state of decoder
    decoder_init_state = tf.split(
        representation,
        2 * layers if bidirectional_decoder else layers,
        axis=-1,
        name="decoder_init_state",
    )
    # Decoder input is reversed and shifted input sequence
    targets = trans[:0:-1]
    dec_inputs = tf.pad(
        targets, [[1, 0], [0, 0], [0, 0]], name="decoder_input_sequence"
    )

    # Make decoder layers and init with output from fully connected layer
    decoder = _make_rnn(units, layers, bidirectional_decoder, name="decoder")
    decoder_output = decoder(dec_inputs, initial_state=decoder_init_state)[0]

    n_features = input_shape[-1]
    reconstruction = Dense(n_features, activation="tanh", name="reconstruction")(
        decoder_output
    )

    reconstruction = tf.transpose(reconstruction, [1, 0, 2])

    model = Model(inputs=inputs, outputs=[reconstruction, representation])
    return model



import logging
import warnings
from abc import ABC, abstractmethod
from itertools import chain, tee
from typing import ClassVar, Dict, Iterable, List, Optional, Type, Union, cast

import librosa
import numpy as np
from omegaconf import OmegaConf

from ertk.config import ERTKConfig
from ertk.utils import PathOrStr, batch_iterable


class InstanceProcessor(ABC):
    """An instance processor.

    Parameters
    ----------
    config: ERTKConfig
        The configuration for this AudioClipProcessor.
    """

    config: ERTKConfig
    """The configuration for this AudioClipProcessor."""
    logger: logging.Logger

    _config_type: ClassVar[Type[ERTKConfig]]
    _friendly_name: ClassVar[str]
    _registry: ClassVar[Dict[str, Type["InstanceProcessor"]]] = {}

    def __init__(self, config: ERTKConfig) -> None:
        self.config = config
        cls_name = f"{InstanceProcessor.__module__}.{InstanceProcessor.__name__}"
        self.logger = logging.getLogger(cls_name)

    def __init_subclass__(
        cls, fname: str = None, config: Type[ERTKConfig] = None
    ) -> None:
        cls._registry = {}
        if fname and config:
            cls._friendly_name = fname
            cls._config_type = config
            for t in cls.mro()[1:-1]:
                t = cast(Type[InstanceProcessor], t)  # For MyPy
                if not hasattr(t, "_registry"):
                    continue
                if fname in t._registry:
                    prev_cls = t._registry[fname]
                    msg = f"Name {fname} already registered with class {prev_cls}."
                    if prev_cls is cls:
                        warnings.warn(msg)
                    else:
                        raise KeyError(msg)
                t._registry[fname] = cls

    @classmethod
    def get_processor_class(cls, name: str) -> Type["InstanceProcessor"]:
        return cls._registry[name]

    @classmethod
    def make_processor(cls, name: str, config: ERTKConfig) -> "InstanceProcessor":
        return cls.get_processor_class(name)(config)

    @classmethod
    def get_config_type(cls) -> Type[ERTKConfig]:
        return cls._config_type

    @classmethod
    def valid_preprocessors(cls) -> List[str]:
        return list(cls._registry)

    @classmethod
    def get_default_config(cls):
        return OmegaConf.structured(cls._config_type)

    @abstractmethod
    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Process a single audio clip.

        Parameters
        ----------
        x: np.ndarray
            The audio data to process.

        Returns
        -------
        result: np.ndarray
            The processed instance.
        """
        raise NotImplementedError()

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        """Process a batch of instances. By default this simply calls
        `process_instance()` on each instance in the batch.

        Parameters
        ----------
        batch: np.ndarray or iterable of arrays
            The batch of instances to process.

        Returns
        -------
        processed: list of np.ndarray
            A list of processed instances.
        """
        return [self.process_instance(x, **kwargs) for x in batch]

    def process_all(
        self, xs: Union[Iterable[np.ndarray], np.ndarray], batch_size: int, **kwargs
    ) -> Iterable[np.ndarray]:
        """Process all instances in batches.

        Parameters
        ----------
        xs: iterable of np.ndarray
            The data from instances to process.
        batch_size: int
            Batch size.

        Returns
        -------
        processed: iterable of np.ndarray
            A generator that yields each processed instance in order.
        """
        if batch_size == 1:
            for x in xs:
                yield self.process_instance(x, **kwargs)
        elif batch_size > 1:
            yield from chain.from_iterable(
                self.process_batch(filter(lambda x: x is not None, b), **kwargs)
                for b in batch_iterable(xs, batch_size)
            )
        elif batch_size < 0:
            yield from self.process_batch(xs, **kwargs)
        else:
            raise ValueError("Batch size cannot be 0.")


class AudioClipProcessor(InstanceProcessor):
    """Processes raw audio data."""

    def process_file(self, path: PathOrStr, sr: Optional[float] = None) -> np.ndarray:
        """Process individual audio file.

        Parameters
        ----------
        path: pathlike or str
            The path to an audio file.
        sr: float, optional
            Target sample rate. If not given, then each audio file is
            loaded with its native sample rate. Otherwise audio data is
            resampled to this sample rate.

        Returns
        -------
        processed: np.ndarray
            Processed instance.
        """
        audio, _sr = librosa.load(path, sr=sr, mono=True)
        return self.process_instance(audio, sr=_sr)

    def process_files(
        self, paths: Iterable[PathOrStr], batch_size: int, sr: Optional[float] = None
    ) -> Iterable[np.ndarray]:
        """Process a set of files.

        Parameters
        ----------
        paths: iterables of path or iterable of str
            Paths to audio files.
        batch_size: int
            Batch size. Note that if `batch_size != 1` then either `sr`
            must be given, or all files must have the same native sample
            rate.
        sr: float, optional
            Target sample rate. If not given, then each audio file is
            loaded with its native sample rate. Otherwise audio data is
            resampled to this sample rate.

        Returns
        -------
        processed: iterable of np.ndarray
            Iterable of processed instances.
        """
        if batch_size == 1:
            # Special case required for variable sample rate
            for path in paths:
                yield self.process_file(path, sr=sr)
        else:
            paths, tmp = tee(paths)
            _sr = librosa.load(next(tmp), sr=sr, duration=0)[1]
            audios = (librosa.load(path, sr=_sr, mono=True)[0] for path in paths)
            yield from self.process_all(audios, batch_size=batch_size, sr=_sr)


class FeatureExtractor(InstanceProcessor):
    @property
    @abstractmethod
    def dim(self) -> int:
        """The dimensionality of the extracted features."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_sequence(self) -> bool:
        """Whether this FeatureExtractor yields sequence features."""
        raise NotImplementedError()
